"""
Note:
   Creation of loggers is done in _create_default_callbacks which check for inheritance from Callback class.

"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ray.air.constants import EXPR_PROGRESS_FILE
from ray.tune.logger import CSVLoggerCallback

from ray_utilities.callbacks.tuner._file_logger_fork_mixin import FileLoggerForkMixin
from ray_utilities.callbacks.tuner.new_style_logger_callback import NewStyleLoggerCallback
from ray_utilities.postprocessing import remove_videos

if TYPE_CHECKING:
    from pathlib import Path

    from ray.tune.experiment.trial import Trial

    from ray_utilities.typing.metrics import AnyLogMetricsDict


logger = logging.getLogger(__name__)


class AdvCSVLoggerCallback(NewStyleLoggerCallback, FileLoggerForkMixin, CSVLoggerCallback):
    """Logs trial results in CSV format.

    Prevents logging of videos (keys in :const:`DEFAULT_VIDEO_DICT_KEYS`) even if they are present
    at the first iteration.

    When a trial is forked (has ``FORK_FROM`` in config), creates a new CSV file
    and tries to copy the data from the parent trial.
    """

    def _get_file_extension(self) -> str:
        return "csv"

    def _get_file_base_name(self) -> str:
        return "progress"

    def _get_default_file_name(self) -> str:
        return EXPR_PROGRESS_FILE

    def _setup_file_handle(self, trial: Trial, local_file_path: Path) -> None:
        """Open the CSV file handle and set CSV-specific state."""
        self._trial_csv[trial] = None  # need to set key # pyright: ignore[reportArgumentType]
        if local_file_path.exists():
            # Check if header needs to be written
            write_header_again = local_file_path.stat().st_size == 0
            # Note: metrics might have changed when loading from checkpoint
            if trial.config.get("cli_args", {}).get("from_checkpoint") or trial.config.get("from_checkpoint"):
                write_header_again = True
            self._trial_continue[trial] = not write_header_again
        else:  # For now this is not called as we enter after a sync from parent
            self._restore_from_remote(local_file_path.name, trial)
            self._trial_continue[trial] = False
        self._trial_files[trial] = local_file_path.open("at")

    def _handle_missing_parent_file(self, trial: Trial, local_file_path: Path) -> None:
        """Handle missing parent file for CSV logger."""
        self._trial_continue[trial] = False
        logger.warning(
            "Trial %s forked but found no logfile for parent, starting fresh .csv log file: %s",
            trial.trial_id,
            local_file_path,
        )

    def log_trial_result(self, iteration: int, trial: "Trial", result: AnyLogMetricsDict):  # pyright: ignore[reportIncompatibleMethodOverride]
        if trial not in self._trial_csv:
            # Keys are permanently set; remove videos from the first iteration.
            # Therefore also need eval metric in first iteration
            result = remove_videos(result)

        super().log_trial_result(
            iteration,
            trial,
            result,
        )


if TYPE_CHECKING:  # Check ABC
    AdvCSVLoggerCallback()



from __future__ import annotations

from functools import partial
import logging
import sys
from typing import TYPE_CHECKING, Any, List, Optional, cast

from ray.rllib.utils.annotations import override
from ray.tune.callback import Callback
from ray.tune.experiment import Trial
from ray.tune.result import SHOULD_CHECKPOINT
from typing_extensions import deprecated, Self

from ray_utilities.training.helpers import get_current_step

if TYPE_CHECKING:
    from collections.abc import Callable

    from ray.tune.experiment import Trial

    from ray_utilities.typing.algorithm_return import StrictAlgorithmReturnData
    from ray_utilities.typing.metrics import LogMetricsDict

_logger = logging.getLogger(__name__)


TUNE_RESULT_IS_A_COPY = True  # NOTE: This is not a dynamic or checked value.
"""
Tuner does not allow to modify the result dict as it is a copy. As long as this is True
use the callback on an Algorithm. Or the HACK in the DefaultTrainable class to trigger
checkpointing.

See Also:
    - https://github.com/ray-project/ray/pull/55527
"""


# todo: do not derive from RLlibCallback when tuner checkpoint is actually working.
# Need workaround to set `SHOULD_CHECKPOINT` in the actual result dict and not on a copy of
# on_trial_result
@deprecated("Do not use as long as tune passes only a copy of the result dict.", stacklevel=2)
class MetricCheckpointer(Callback):
    """Callbacks that adds ``SHOULD_CHECKPOINT`` to results if a metric condition is met."""

    condition: Optional[Callable[[dict], bool]] = None
    _last_checkpoint_step = -1

    def __init__(self, metric_name: Optional[str] = None, condition: Optional[Callable[[dict], bool]] = None) -> None:
        if self.condition is None and condition is None:
            raise ValueError(
                "Condition must be provided for MetricCheckpointer. Either as class variable or in constructor."
            )
        if self.condition is not None and condition is not None:
            _logger.warning("Both class variable and constructor condition provided. Using constructor condition.")
        super().__init__()
        self.metric_name = metric_name or "Unknown"
        self.condition = self.condition or condition
        assert self.condition
        self._last_checkpoint_iteration = -1
        self._last_checkpoint_value = None

    def _set_checkpoint(self, result: StrictAlgorithmReturnData | LogMetricsDict, trial: Trial | None = None) -> None:
        iteration = result["training_iteration"]
        current_step = get_current_step(result)
        # config available in trial.config
        if self.condition(cast("dict[str, Any]", result)):  # pyright: ignore[reportOptionalCall]
            self._last_checkpoint_iteration = iteration
            self._last_checkpoint_value = result.get(self.metric_name, None)
            self._last_checkpoint_step = current_step
            result[SHOULD_CHECKPOINT] = True  # XXX # NOTE: That this is a copy and does NOT WORK
            _logger.info(
                "Checkpointing trial %s at iteration %s, step %d with: metric '%s' = %s%s",
                trial.trial_id if trial else "",
                iteration,
                current_step,
                self.metric_name,
                self._last_checkpoint_value,
                (
                    ". NOTE: This is only a logging message and does not confirm the checkpoint creation"
                    if TUNE_RESULT_IS_A_COPY
                    else ""
                ),
            )

    @override(Callback)
    def on_trial_result(
        self,
        iteration: int,
        trials: List["Trial"],
        trial: "Trial",
        result: dict,
        **info,
    ):
        """Called after receiving a result from a trial.

        The search algorithm and scheduler are notified before this
        hook is called.

        Arguments:
            iteration: Number of iterations of the tuning loop.
            trials: List of trials.
            trial: Trial that just sent a result.
            result: Result that the trial sent.
            **info: Kwargs dict for forward compatibility.
        """
        self._set_checkpoint(
            result,  # pyright: ignore[reportArgumentType]
            trial,
        )


@deprecated("Do not use as long as tune passes only a copy of the result dict.", stacklevel=2)
class StepCheckpointer(MetricCheckpointer):  # type: ignore
    """Checkpoints trials based on a specific metric condition."""

    def _condition(self, result: StrictAlgorithmReturnData | LogMetricsDict | dict) -> bool:
        steps_since_last_checkpoint = get_current_step(result) - self._last_checkpoint_step  # pyright: ignore[reportArgumentType]
        # _logger.debug(
        #    "StepCheckpointer: steps since last checkpoint: %d, frequency: %d",
        #    steps_since_last_checkpoint,
        #    self._checkpoint_frequency,
        # )
        return steps_since_last_checkpoint >= self._checkpoint_frequency

    def __init__(self, checkpoint_frequency: int = 50_000) -> None:
        if checkpoint_frequency == 0:
            _logger.info("Checkpoint frequency is set to 0, disabling step checkpointing.")
            checkpoint_frequency = sys.maxsize
        self._checkpoint_frequency = checkpoint_frequency
        super().__init__("current_step", self._condition)

    @classmethod
    def make_callback_class(cls, *, checkpoint_frequency, **kwargs) -> type[Self]:
        return partial(cls, checkpoint_frequency=checkpoint_frequency, **kwargs)  # pyright: ignore[reportReturnType]



"""
Type definitions for postprocessing metrics.

For algorithm return data, see `ray_utilities.typing.algorithm_return`
"""

# pyright: enableExperimentalFeatures=true
from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Any, Literal, TypeGuard

from typing_extensions import Never, NotRequired, Required, TypedDict

from .algorithm_return import EvaluationResultsDict, _EvaluationNoDiscreteDict
from .common import BaseEnvRunnersResultsDict, VideoTypes

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from ray.rllib.utils.typing import AgentID, ModuleID

__all__ = [
    "LogMetricsDict",
]


class VideoMetricsDict(TypedDict, closed=True):
    video: VideoTypes.LogVideoTypes
    """
    A 5D numpy array representing a video; or a string pointing to a video file to upload.

    Should be a list of a 5D numpy video array representing a video.
    """
    reward: float
    video_path: NotRequired[str]
    """If a video file already exists for re-use, this is the path to the video file."""


class _WarnVideosToEnvRunners(TypedDict):
    episode_videos_best: NotRequired[Annotated[Never, "needs to be in env_runners"]]
    episode_videos_worst: NotRequired[Annotated[Never, "needs to be in env_runners"]]


class _LogMetricsEnvRunnersResultsDict(BaseEnvRunnersResultsDict):
    """Environment runner results optimized for logging metrics.

    Extends the base type with additional optional fields used in logging.
    Most fields are NotRequired in metrics context as they may not always be available.

    See Also:
        :class:`BaseEnvRunnersResultsDict`: Common base type
        :data:`ray.rllib.utils.metrics.ENV_RUNNER_RESULTS`: Ray RLlib env runner metrics
    """

    episode_return_max: NotRequired[float]
    episode_return_min: NotRequired[float]
    num_env_steps_sampled_lifetime: NotRequired[int]
    num_env_steps_sampled: NotRequired[int]
    num_env_steps_passed_to_learner: NotRequired[int]
    """Custom key for exact current_steps added by exact_sampling_callback"""
    num_env_steps_passed_to_learner_lifetime: NotRequired[int]
    """Custom key for exact current_steps added by exact_sampling_callback"""
    episode_duration_sec_mean: NotRequired[float]
    num_module_steps_sampled: NotRequired[dict[ModuleID, int]]
    num_module_steps_sampled_lifetime: NotRequired[dict[ModuleID, int]]
    num_agent_steps_sampled: NotRequired[dict[AgentID, int]]
    num_agent_steps_sampled_lifetime: NotRequired[dict[AgentID, int]]


class _LogMetricsEvalEnvRunnersResultsDict(_LogMetricsEnvRunnersResultsDict, total=False):
    """Environment runner results for evaluation with video support."""

    episode_videos_best: VideoTypes.LogVideoTypes | VideoMetricsDict
    episode_videos_worst: VideoTypes.LogVideoTypes | VideoMetricsDict


class _LogMetricsEvaluationResultsWithoutDiscreteDict(_EvaluationNoDiscreteDict, _WarnVideosToEnvRunners):
    env_runners: Required[_LogMetricsEvalEnvRunnersResultsDict]  # pyright: ignore[reportIncompatibleVariableOverride]
    discrete: NotRequired[Never]


class _LogMetricsEvaluationResultsDict(EvaluationResultsDict, _WarnVideosToEnvRunners):
    env_runners: _LogMetricsEvalEnvRunnersResultsDict  # pyright: ignore[reportIncompatibleVariableOverride]
    discrete: NotRequired[_LogMetricsEvaluationResultsWithoutDiscreteDict]  # pyright: ignore[reportIncompatibleVariableOverride]


class _LogMetricsBase(TypedDict):
    training_iteration: int
    """The number of times train.report() has been called"""

    current_step: int
    """
    The current step in the training process, usually the number of environment steps sampled.

    For exact sampling use:

        - "learners/(__all_modules__ | default_policy)/num_env_steps_passed_to_learner_lifetime"
            Requires: ``RemoveMaskedSamplesConnector`` (+ ``exact_sampling_callback`` at best)
        - "env_runners/num_env_steps_sampled_lifetime"
            Requires: ``exact_sampling_callback``

    Otherwise use::

            env_runners/num_env_steps_sampled_lifetime
    """

    done: bool
    timers: NotRequired[dict[str, dict[str, Any] | Any]]
    learners: NotRequired[dict[ModuleID | Literal["__all_modules__"], dict[str, Any] | Any]]

    fault_tolerance: NotRequired[Any]
    env_runner_group: NotRequired[Any]
    num_env_steps_sampled_lifetime: NotRequired[int]
    num_env_steps_sampled_lifetime_throughput: NotRequired[float]
    """Mean time in seconds between two logging calls to num_env_steps_sampled_lifetime"""

    should_checkpoint: NotRequired[bool]
    """If True, the tuner should checkpoint this step."""

    batch_size: NotRequired[int]
    """Current train_batch_size_per_learner. Should be logged in experiments were it can change."""

    num_training_step_calls_per_iteration: NotRequired[int]
    """How training_steps was called between two train.report() calls."""

    config: NotRequired[dict[str, Any]]
    """Algorithm config used for this training step."""


class LogMetricsDict(_LogMetricsBase):
    """Stays true to RLlib's naming."""

    env_runners: _LogMetricsEnvRunnersResultsDict
    evaluation: _LogMetricsEvaluationResultsDict


class AutoExtendedLogMetricsDict(LogMetricsDict):
    """
    Auto filled in keys after train.report.

    Use this in Callbacks

    See Also:
        - https://docs.ray.io/en/latest/tune/tutorials/tune-metrics.html#tune-autofilled-metrics
        - Trial.last_result
    """

    done: bool
    training_iteration: int  # auto filled in
    """The number of times train.report() has been called"""
    trial_id: int | str  # auto filled in


FlatLogMetricsDict = TypedDict(
    "FlatLogMetricsDict",
    {
        "training_iteration": int,
        "env_runners/episode_return_mean": float,
        "evaluation/env_runners/episode_return_mean": float,
        "evaluation/env_runners/episode_videos_best": NotRequired["str | NDArray"],
        "evaluation/env_runners/episode_videos_best/reward": NotRequired[float],
        "evaluation/env_runners/episode_videos_best/video": NotRequired["NDArray"],
        "evaluation/env_runners/episode_videos_best/video_path": NotRequired["str"],
        "evaluation/env_runners/episode_videos_worst": NotRequired["NDArray | str"],
        "evaluation/env_runners/episode_videos_worst/reward": NotRequired[float],
        "evaluation/env_runners/episode_videos_worst/video": NotRequired["NDArray"],
        "evaluation/env_runners/episode_videos_worst/video_path": NotRequired["str"],
        "evaluation/discrete/env_runners/episode_return_mean": float,
        "evaluation/discrete/env_runners/episode_videos_best": NotRequired["str | NDArray"],
        "evaluation/discrete/env_runners/episode_videos_best/reward": NotRequired[float],
        "evaluation/discrete/env_runners/episode_videos_best/video": NotRequired["NDArray"],
        "evaluation/discrete/env_runners/episode_videos_best/video_path": NotRequired["str"],
        "evaluation/discrete/env_runners/episode_videos_worst": NotRequired["str | NDArray"],
        "evaluation/discrete/env_runners/episode_videos_worst/reward": NotRequired[float],
        "evaluation/discrete/env_runners/episode_videos_worst/video": NotRequired["NDArray"],
        "evaluation/discrete/env_runners/episode_videos_worst/video_path": NotRequired["str"],
    },
    closed=False,
)


class _NewLogMetricsEvaluationResultsWithoutDiscreteDict(_LogMetricsEvalEnvRunnersResultsDict):
    discrete: NotRequired[Never]


class _NewLogMetricsEvaluationResultsDict(_LogMetricsEvalEnvRunnersResultsDict):
    discrete: NotRequired[_NewLogMetricsEvaluationResultsWithoutDiscreteDict]


class NewLogMetricsDict(_LogMetricsBase):
    """
    Changes:
        env_runners -> training
        evaluation.env_runners -> evaluation

        if only "__all_modules__" and "default_policy" are present in learners,
        merges them.
    """

    training: _LogMetricsEnvRunnersResultsDict
    evaluation: _NewLogMetricsEvaluationResultsDict


AnyLogMetricsDict = LogMetricsDict | NewLogMetricsDict


class NewAutoExtendedLogMetricsDict(NewLogMetricsDict):
    """
    Auto filled in keys after train.report.

    Use this in Callbacks

    See Also:
        - https://docs.ray.io/en/latest/tune/tutorials/tune-metrics.html#tune-autofilled-metrics
        - Trial.last_result
    """

    done: bool
    training_iteration: int  # auto filled in
    """The number of times train.report() has been called"""
    trial_id: int | str  # auto filled in


AnyAutoExtendedLogMetricsDict = AutoExtendedLogMetricsDict | NewAutoExtendedLogMetricsDict

NewFlatLogMetricsDict = TypedDict(
    "NewFlatLogMetricsDict",
    {
        "training_iteration": int,
        "training/episode_return_mean": float,
        "evaluation/episode_return_mean": float,
        "evaluation/episode_videos_best": NotRequired["str | NDArray"],
        "evaluation/episode_videos_best/reward": NotRequired[float],
        "evaluation/episode_videos_best/video": NotRequired["NDArray"],
        "evaluation/episode_videos_best/video_path": NotRequired["str"],
        "evaluation/episode_videos_worst": NotRequired["NDArray | str"],
        "evaluation/episode_videos_worst/reward": NotRequired[float],
        "evaluation/episode_videos_worst/video": NotRequired["NDArray"],
        "evaluation/episode_videos_worst/video_path": NotRequired["str"],
        "evaluation/discrete/episode_return_mean": float,
        "evaluation/discrete/episode_videos_best": NotRequired["str | NDArray"],
        "evaluation/discrete/episode_videos_best/reward": NotRequired[float],
        "evaluation/discrete/episode_videos_best/video": NotRequired["NDArray"],
        "evaluation/discrete/episode_videos_best/video_path": NotRequired["str"],
        "evaluation/discrete/episode_videos_worst": NotRequired["str | NDArray"],
        "evaluation/discrete/episode_videos_worst/reward": NotRequired[float],
        "evaluation/discrete/episode_videos_worst/video": NotRequired["NDArray"],
        "evaluation/discrete/episode_videos_worst/video_path": NotRequired["str"],
    },
    closed=False,
)

AnyFlatLogMetricsDict = FlatLogMetricsDict | NewFlatLogMetricsDict

# region TypeGuards


def has_video_key(
    dir: dict | _LogMetricsEvalEnvRunnersResultsDict, video_key: Literal["episode_videos_best", "episode_videos_worst"]
) -> TypeGuard[_LogMetricsEvalEnvRunnersResultsDict]:
    return video_key in dir


# endregion TypeGuards



"""Miscellaneous utilities for Ray RLlib workflows.

Provides various utility functions for working with Ray Tune experiments,
progress bars, and data structures. Includes functions for trial naming,
trainable introspection, dictionary operations, and error handling.
"""

from __future__ import annotations

import datetime
import functools
import logging
import os
import re
import sys
from typing import TYPE_CHECKING, Any, Mapping, Optional, TypeVar

import base62
from ray.experimental import tqdm_ray
from ray.tune.error import TuneError
from ray.tune.result_grid import ResultGrid
from tqdm import tqdm
from typing_extensions import Iterable, TypeIs

from ray_utilities.constants import (
    DEFAULT_EVAL_METRIC,
    EVAL_METRIC_RETURN_MEAN,
    NEW_LOG_EVAL_METRIC,
    RAY_UTILITIES_INITIALIZATION_TIMESTAMP,
    RE_PARSE_FORK_FROM,
    RUN_ID,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from ray.tune.experiment import Trial

    from ray_utilities.typing import ForkFromData

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup

_T = TypeVar("_T")

_logger = logging.getLogger(__name__)

RE_GET_TRIAL_ID = re.compile(r"id=(?P<trial_id>(?P<trial_id_part1>[a-zA-Z0-9]{5,6})(?:_(?P<trial_number>[0-9]{5}))?)")
"""Regex pattern to extract the trial ID from checkpoint paths.

This pattern assumes the trial ID is in the format 'id=<part1>[_<trial_number>]',
with the trial number being optional. The length of each block is not validated.

Example:
    >>> match = RE_TRIAL_ID_FROM_CHECKPOINT.search("path/to/checkpoint/id=abc123_000001")
    >>> match.group("trial_id") if match else None
    'abc123'
"""


def extract_trial_id_from_checkpoint(ckpt_path: str) -> str | None:
    """Extract the trial ID from a checkpoint path.

    This function uses a regex pattern to extract the trial ID from a given
    checkpoint path. The expected format is 'id=<part1>[_<trial_number>]',
    where the trial number is optional.

    Args:
        ckpt_path: The checkpoint path string to extract the trial ID from.

    Note:
        As a fallback, the function also checks for an older format without the 'id=' prefix.
        In this case it looks for a pattern like '<part1>_<trial_number>'.
        Where trial_number is exactly 5 digits and part1 is at least 5 alphanumeric characters.

    Returns:
        The extracted trial ID as a string, or ``None`` if no valid ID is found
    """
    # TODO possibly, trial id might contain _forkof_/fork_from in the future.
    match = RE_GET_TRIAL_ID.search(ckpt_path)
    # get id of run
    if match:
        return match.groupdict()["trial_id"]
    # Deprecated:
    # possible old format without id=
    match = re.search(r"(?:id=)?(?P<trial_id>[a-zA-Z0-9]{5,}_[0-9]{5})", ckpt_path)
    if match:
        return match.groupdict()["trial_id"]
    return None


def parse_fork_from(fork_from: str) -> tuple[str, int | None] | None:
    """Parse a forked trial identifier into its components.

    This function takes a forked trial identifier string and splits it into
    the original trial ID and the step at which the fork occurred. The expected
    format is '<trial_id>?_step=<step>', where the step is optional.

    Args:
        fork_from: The forked trial identifier string to parse.
            Example: "abc123?_step=10" or "abc123"
    Returns:
        A tuple containing the trial ID as a string and the step as an integer
        or ``None`` if the step is not specified.
        Or None if the input string does not match the expected format.
    """
    # Note: only used for wandb currently, possibly move there
    # NOTE: could also just do split("?_step=") here
    match = RE_PARSE_FORK_FROM.match(fork_from)
    if not match:
        return None
    trial_id = match.group("fork_id")
    step_str = match.group("fork_step")
    step = int(step_str) if step_str is not None else None
    return trial_id, step


def trial_name_creator(trial: Trial) -> str:
    """Create a descriptive name for a Ray Tune trial.

    Generates a human-readable trial name that includes the trainable name,
    environment, module, start time, and trial ID. Optionally includes
    checkpoint information if the trial was restored from a checkpoint.

    Args:
        trial: The :class:`ray.tune.experiment.Trial` object to create a name for.

    Returns:
        A formatted string containing trial information, with fields separated by underscores.
        Format: ``<setup_cls>_<trainable_name>_<env>_<module>_<start_time>_id=<trial_id>``
        with optional ``[_from_checkpoint=<checkpoint_id>]`` suffix.

    Example:
        >>> # For a PPO trial on CartPole started at 2023-01-01 12:00
        >>> trial_name_creator(trial)
        'PPO_CartPole-v1_ppo_2023-01-01_12:00_id=abc123_456'
    """
    start_time = datetime.datetime.fromtimestamp(
        trial.run_metadata.start_time or RAY_UTILITIES_INITIALIZATION_TIMESTAMP
    )
    start_time_str = start_time.strftime("%Y-%m-%d_%H-%M")
    module = trial.config.get("module", None)
    if module is None and "cli_args" in trial.config:
        module = trial.config["cli_args"]["agent_type"]
    fields = [
        trial.config["env"],
        trial.trainable_name,
        module,
        "id=" + trial.trial_id,
        start_time_str,
    ]
    if "cli_args" in trial.config and trial.config["cli_args"]["from_checkpoint"]:
        match = RE_GET_TRIAL_ID.match(trial.config["cli_args"]["from_checkpoint"])
        if match:
            fields.append("from_checkpoint=" + match.group("trial_id"))
    setup_cls = trial.config.get("setup_cls", None)
    if setup_cls is not None:
        fields.insert(0, setup_cls)
    return "_".join(fields)


_NOT_FOUND = object()


def _is_key(key: str):
    return key.startswith("<") and key.endswith(">")


def _format_key(key: str) -> str:
    assert key[0] == "<" and key[-1] == ">", "Key must be wrapped in angle brackets."
    key = key[1:-1]  # remove angle brackets
    if key == "batch_size":
        return "train_batch_size_per_learner"
    return key


def extend_trial_name(
    insert: Iterable[str] = (), *, prepend: Iterable[str] = (), append: Iterable[str] = ()
) -> Callable[[Trial], str]:
    """
    Inserts strings or values from the trials config into the trial name.

    Values to be extracted from the config must be wrapped in angle brackets,
    e.g. "<my_param>". The function returns a new trial name creator that can be
    used in place of the default one.

    Args:
        insert: Iterable of strings or config keys to insert before the "_id=" part.
        prepend: Iterable of strings or config keys to prepend at the start.
        append: Iterable of strings or config keys to append at the end.

    Example:
        name_creator = extend_trial_name(insert=["<param1>"], prepend=["NEW"], append=["<param2>"])
        # This will create trial names like "NEW_<param1>_..._id=..._<param2>"

    Hint:
        For `train_batch_size_per_learner`, you can use the shorthand `<batch_size>` instead of the full key.

    Returns:
        A callable that takes a :class:`ray.tune.experiment.Trial` and returns
        a modified trial name string with the specified insertions.
    """
    if isinstance(insert, str):
        insert = (insert,)
    if isinstance(prepend, str):
        prepend = (prepend,)
    if isinstance(append, str):
        append = (append,)

    def extended_trial_name_creator(trial: Trial) -> str:
        base = trial_name_creator(trial)

        start, end = base.split("_id=")
        for key in insert:
            if _is_key(key):
                value = trial.config.get(_format_key(key), _NOT_FOUND)
                if value is _NOT_FOUND:
                    _logger.warning("Key %s not found in trial config, skipping insertion into trial name.", key)
                    continue
                start += f"_{key[1:-1]}={value}"
            else:
                start += f"_{key}"
        for key in prepend:
            if _is_key(key):
                value = trial.config.get(_format_key(key), _NOT_FOUND)
                if value is _NOT_FOUND:
                    _logger.warning("Key %s not found in trial config, skipping prepending into trial name.", key)
                    continue
                start = f"{key[1:-1]}={value}_" + start
            else:
                start = f"{key}_" + start
        for key in append:
            if _is_key(key):
                value = trial.config.get(_format_key(key), _NOT_FOUND)
                if value is _NOT_FOUND:
                    _logger.warning("Key %s not found in trial config, skipping appending into trial name.", key)
                    continue
                end += f"_{key[1:-1]}={value}"
            else:
                end += f"_{key}"
        return start + "_id=" + end

    return extended_trial_name_creator


def get_trainable_name(trainable: Callable) -> str:
    """Extract the original name from a potentially wrapped trainable function.

    Unwraps :func:`functools.partial` objects and functions with ``__wrapped__``
    attributes to find the original function name. This is useful for identifying
    the underlying trainable when it has been decorated or partially applied.

    Args:
        trainable: A callable that may be wrapped with decorators or partial application.

    Returns:
        The ``__name__`` attribute of the unwrapped function.

    Example:
        >>> import functools
        >>> def my_trainable():
        ...     pass
        >>> wrapped = functools.partial(my_trainable, arg=1)
        >>> get_trainable_name(wrapped)
        'my_trainable'
    """
    last = None
    while last != trainable:
        last = trainable
        while isinstance(trainable, functools.partial):
            trainable = trainable.func
        while hasattr(trainable, "__wrapped__"):
            trainable = trainable.__wrapped__  # type: ignore[attr-defined]
    return trainable.__name__


def _make_non_fork_experiment_key(trial: Trial) -> str:
    trial_base, *trial_number = trial.trial_id.split("_")
    if len(trial_number) > 1:
        _logger.warning(
            "Unexpected trial_id format '%s'. Expected format '<id>_<number>'.",
            trial.trial_id,
        )
    if trial_number:
        trial_number = "C" + "".join(trial_number).replace("000", "")
    else:  # empty list
        trial_number = ""
    base_key = f"{RUN_ID}X{trial_base}{trial_number}".replace("_", "")
    # Pad at the end with Z to be at least 32 characters long.
    # Use uppercase letters as trial_id is lowercase alphanumeric only.
    base_key = f"{base_key:Z<32}"
    return base_key


def _make_fork_experiment_key(base_key: str, fork_data: ForkFromData) -> str:
    base_key = base_key.rstrip("Z")
    parent_id = fork_data["parent_id"]
    # Prefer training iteration for experiment key (stable across frameworks)
    ft = fork_data.get("parent_time")
    # ft is a NamedTuple[time_attr, time]; only use numeric time
    parent_iteration = ft[1] if ft[0] == "current_step" else fork_data["parent_training_iteration"]

    fork_base, *fork_number = parent_id.split("_")
    if fork_number:
        fork_number = "C" + "".join(fork_number).replace("000", "")
    else:
        fork_number = ""

    if parent_iteration is None:
        iteration_data = "NONE"  # rare chance that NONE is actually encoded
        _logger.warning("parent_iteration is None, using 'NONE' in experiment key.")
    else:
        iteration_data = base62.encode(parent_iteration)
    return f"{base_key}F{fork_base}{fork_number}S{iteration_data:0>4}".replace("_", "")


def make_experiment_key(trial: Trial, fork_data: Optional[ForkFromData] = None) -> str:
    """
    Build a unique experiment key for a trial, making use of the :attr:`RUN_ID`,
    :attr:`~ray.tune.experiment.Trial.trial_id <Trial.trial_id>`, and optional fork information.

    It has the format of two forms:
        - If not forked: "<RUN_ID of 21 chars>X<trial_id>[Z* up to 32 chars in total]"
          it is prolonged by ``Z`` to be at least 32 chars long.
        - If forked: "<RUN_ID of 21 chars>X<trial_id>F<trial_id>S<step>"
          The <step> is in base62 encoded and right aligned to 4 characters with leading zeros.

        Both have in common that the potential underscore in the trial ID is replaced by C and all "000" in the trials
        count part of the trial ID are removed, e.g. ``abcd0001_00005`` becomes ``abcd0001C05``.

    References:
        - :attr:`RUN_ID <ray_utilities.constants.RUN_ID>`: A unique identifier for the current execution.
        - :attr:`~ray.tune.experiment.Trial.trial_id <Trial.trial_id>`: The unique ID of the trial.

    Note:
        The resulting key underscores replaced by C. For compatibility it should only contain
        numbers and letters.
        For comet it must be between 32 and 50 characters long. This is not enforced here.

        For comet support, to not exceed 50 characters, only 99 parallel trials are supported, e.g. <trial_id>_00099.
        The highest supported forking step is 14_776_335.

        It is assumed that the trial ID contains only lowercase alphanumeric characters and underscores.
    """
    base_key = _make_non_fork_experiment_key(trial)
    if not fork_data:
        return base_key
    return _make_fork_experiment_key(base_key, fork_data)


def is_pbar(pbar: Iterable[_T]) -> TypeIs[tqdm_ray.tqdm | tqdm[_T]]:
    """Type guard to check if an iterable is a tqdm progress bar.

    This function serves as a :class:`typing_extensions.TypeIs` guard to narrow
    the type of an iterable to either :class:`ray.experimental.tqdm_ray.tqdm` or
    :class:`tqdm.tqdm`.

    Args:
        pbar: An iterable that might be a progress bar.

    Returns:
        ``True`` if the object is a tqdm or tqdm_ray progress bar, ``False`` otherwise.

    Example:
        >>> from tqdm import tqdm
        >>> progress = tqdm(range(10))
        >>> if is_pbar(progress):
        ...     # Type checker now knows progress is a tqdm object
        ...     progress.set_description("Processing")
    """
    return isinstance(pbar, (tqdm_ray.tqdm, tqdm))


def deep_update(mapping: dict[str, Any], *updating_mappings: dict[str, Any]) -> dict[str, Any]:
    """Recursively update a dictionary with one or more updating dictionaries.

    This function performs a deep merge of dictionaries, where nested dictionaries
    are recursively merged rather than replaced. Non-dictionary values are overwritten.

    Note:
        This implementation is adapted from `Pydantic's internal utilities
        <https://github.com/pydantic/pydantic/blob/main/pydantic/_internal/_utils.py>`_.

    Args:
        mapping: The base dictionary to update.
        *updating_mappings: One or more dictionaries to merge into the base mapping.

    Returns:
        A new dictionary containing the merged result. The original dictionaries
        are not modified.

    Example:
        >>> base = {"a": {"x": 1, "y": 2}, "b": 3}
        >>> update = {"a": {"y": 20, "z": 30}, "c": 4}
        >>> deep_update(base, update)
        {"a": {"x": 1, "y": 20, "z": 30}, "b": 3, "c": 4}
    """
    updated_mapping = mapping.copy()
    for updating_mapping in updating_mappings:
        for k, v in updating_mapping.items():
            if k in updated_mapping and isinstance(updated_mapping[k], dict) and isinstance(v, dict):
                updated_mapping[k] = deep_update(updated_mapping[k], v)
            else:
                updated_mapping[k] = v
    return updated_mapping


def raise_tune_errors(result: ResultGrid | Sequence[Exception], msg: str = "Errors encountered during tuning") -> None:
    """Raise errors from Ray Tune results as a single ExceptionGroup.

    Processes errors from Ray Tune training results and raises them in a structured way.
    If only one error is present, it's raised directly. Multiple errors are grouped
    using :class:`ExceptionGroup`.

    Args:
        result: Either a :class:`ray.tune.result_grid.ResultGrid` containing errors,
            or a sequence of exceptions to raise.
        msg: Custom message for the ExceptionGroup. Defaults to
            "Errors encountered during tuning".

    Raises:
        Exception: The single error if only one is present.
        ExceptionGroup: Multiple errors grouped together with the provided message.

    Returns:
        None if no errors are found in the ResultGrid.

    Example:
        >>> from ray.tune import ResultGrid
        >>> # Assuming result_grid contains errors from failed trials
        >>> raise_tune_errors(result_grid, "Training failed")
    """
    if isinstance(result, ResultGrid):
        if not result.errors:
            return
        for i, error in enumerate(result.errors):
            if not isinstance(error, BaseException):
                _logger.debug("Error %d is not an exception: %s", i, error)
                exception = TuneError("Error(s) occurred:\n" + str(error))
                result.errors[i] = exception
        if len(result.errors) == 1:
            raise result.errors[0]
        errors = result.errors
    else:
        errors = result
    raise ExceptionGroup(msg, errors)


class AutoInt(int):
    """An integer subclass that represents an automatically determined value.

    This class extends :class:`int` to provide a semantic distinction for values
    that were originally specified as "auto" in command-line arguments or configuration,
    but have been resolved to specific integer values.

    The class maintains the same behavior as a regular integer but can be used
    for type checking and to track the origin of automatically determined values.

    Example:
        >>> value = AutoInt(42)  # Originally "auto", resolved to 42
        >>> isinstance(value, int)  # True
        >>> isinstance(value, AutoInt)  # True
        >>> value + 10  # 52
    """


_new_log_format_used: bool | None = None


def new_log_format_used() -> bool:
    """Check if the new log format is enabled via environment variable.

    This function checks the environment variable
    :envvar:`RAY_UTILITIES_NEW_LOG_FORMAT` to determine if the new logging format
    should be used. The new format changes how metrics are structured in logs that are,
    for example, sent to WandB or Comet ML.

    Returns:
        ``True`` if the environment variable is set to a truthy value (not "0", "false", or "off"),
        ``False`` otherwise.

    Note:
        The result is cached after the first call.
    """
    global _new_log_format_used  # noqa: PLW0603
    if _new_log_format_used is not None:
        return _new_log_format_used
    _new_log_format_used = "RAY_UTILITIES_NEW_LOG_FORMAT" in os.environ and os.environ[
        "RAY_UTILITIES_NEW_LOG_FORMAT"
    ].lower() not in (
        "0",
        "false",
        "off",
    )
    return _new_log_format_used


def resolve_default_eval_metric(eval_metric: str | DEFAULT_EVAL_METRIC | None = None) -> str:
    """Resolve the default evaluation metric based on log format.

    This function determines the evaluation metric key to use
    based on whether the new log format is enabled. If the provided
    `eval_metric` is ``None``, it defaults to either
    :attr:`EVAL_METRIC_RETURN_MEAN <ray_utilities.constants.EVAL_METRIC_RETURN_MEAN>` or
    :attr:`NEW_LOG_EVAL_METRIC <ray_utilities.constants.NEW_LOG_EVAL_METRIC>` depending on
    the log format.

    Args:
        eval_metric: The evaluation metric key to resolve, or ``None`` to use the default.

    Returns:
        The resolved evaluation metric key as a string.

    Attention:
        This function is intended to be used with loggers like the WandB or Comet logger
        where the log metrics are changed for better human interpretation and not
        for other callbacks where the original metric keys are expected.
    """
    if eval_metric is not DEFAULT_EVAL_METRIC and eval_metric is not None:
        return eval_metric
    if new_log_format_used():
        return NEW_LOG_EVAL_METRIC
    return EVAL_METRIC_RETURN_MEAN


def get_value_by_path(tree_dict: Mapping[Any, Any], path: tuple[Any, ...]) -> Any:
    """Extract value from nested dict using a tuple path."""
    value = tree_dict
    for key in path:
        value = value[key]
    return value


def flatten_mapping_with_path(d, path=()) -> list[tuple[tuple[Any, ...], Any]]:
    """Similar to :func:`tree.flatten_with_path` but only for mappings."""
    items = []
    if isinstance(d, Mapping):
        for k, v in d.items():
            items.extend(flatten_mapping_with_path(v, (*path, k)))
    else:
        items.append((path, d))
    return items


# Build a nested dict with lists at the leaves
def build_nested_dict(paths: list[tuple[Any, ...]], values: dict[tuple[Any, ...], list]) -> dict:
    """Build a nested dict from paths and corresponding values."""
    nested = {}
    for path in paths:
        d = nested
        for key in path[:-1]:
            d = d.setdefault(key, {})
        d[path[-1]] = values[path]
    return nested


def flat_dict_to_nested(metrics: dict[str, Any]) -> dict[str, Any | dict[str, Any]]:
    """Convert a flat dictionary with slash-separated keys to a nested dictionary structure.

    This function transforms dictionary keys containing forward slashes into nested
    dictionary structures, useful for organizing Ray Tune/RLlib metrics that are
    typically logged with hierarchical key names.

    Args:
        metrics: A dictionary with potentially slash-separated keys (e.g.,
            ``{"eval/return_mean": 100, "train/loss": 0.5}``).

    Returns:
        A nested dictionary structure where slash-separated keys become nested levels
        (e.g., ``{"eval": {"return_mean": 100}, "train": {"loss": 0.5}}``).

    Example:
        >>> metrics = {
        ...     "train/episode_return_mean": 150.0,
        ...     "eval/env_runner_results/episode_return_mean": 200.0,
        ...     "timesteps_total": 10000,
        ... }
        >>> nested = flat_dict_to_nested(metrics)
        >>> nested["train"]["episode_return_mean"]
        150.0
        >>> nested["eval"]["env_runner_results"]["episode_return_mean"]
        200.0

    Note:
        This is particularly useful when working with Ray Tune's result dictionaries
        which often contain hierarchical metrics with slash-separated key names.
    """
    nested_metrics = metrics.copy()
    for key_orig, v in metrics.items():
        k = key_orig
        subdict = nested_metrics
        while "/" in k:
            parent, k = k.split("/", 1)
            subdict = subdict.setdefault(parent, {})
        subdict[k] = v
        if key_orig != k:
            del nested_metrics[key_orig]
    return nested_metrics



"""Environment creation and configuration utilities for Ray RLlib experiments.

Provides utilities for creating and configuring Gymnasium environments for use with
Ray RLlib algorithms. Includes environment registration, seeding utilities, and
helper functions for environment name parsing.

Key Components:
    - :func:`create_env`: Environment creation with name parsing
    - :func:`parse_env_name`: Environment name shorthand resolution
    - :func:`create_env_for_config`: Environment creation from RLlib configs

Environment Shortcuts:
    - ``"lunar"`` → ``"LunarLander-v2"``
    - ``"cart"`` / ``"cartpole"`` → ``"CartPole-v1"``

Example:
    >>> env = create_env("cart")  # Creates CartPole-v1
    >>> env = create_env("LunarLander-v2")  # Direct environment ID
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import gymnasium as gym
import numpy as np
from ray.tune.registry import register_env
from typing_extensions import deprecated

if TYPE_CHECKING:
    from ray.rllib.algorithms import AlgorithmConfig
    from ray.rllib.env.env_context import EnvContext

__all__ = [
    "create_env",
    "env_creator_with_seed",
    "parse_env_name",
]


env_short_names = {
    "lunar": "LunarLander-v2",
    "cart": "CartPole-v1",
    "cartpole": "CartPole-v1",
}

_logger = logging.getLogger(__name__)


def parse_env_name(name: str) -> str:
    """Parse environment name, converting shortcuts to full Gymnasium environment IDs.

    This function provides convenient shortcuts for commonly used environments,
    allowing users to specify shorter names that are automatically expanded
    to their full Gymnasium environment identifiers.

    Args:
        name: Environment name or shortcut. Can be a full Gymnasium environment ID
            or one of the supported shortcuts.

    Returns:
        Full Gymnasium environment ID. If the input is already a full ID or
        not a recognized shortcut, returns the input unchanged.

    Supported Shortcuts:
        - ``"lunar"`` → ``"LunarLander-v2"``
        - ``"cart"`` → ``"CartPole-v1"``
        - ``"cartpole"`` → ``"CartPole-v1"``

    Example:
        >>> parse_env_name("cart")
        "CartPole-v1"
        >>> parse_env_name("LunarLander-v2")
        "LunarLander-v2"
    """
    return env_short_names.get(name, name)


def create_env(name: str, **kwargs) -> gym.Env:
    """Create a Gymnasium environment with optional name shortcut resolution.

    This function creates Gymnasium environments while supporting convenient
    name shortcuts for commonly used environments. Additional keyword arguments
    are passed directly to the Gymnasium ``make`` function.

    Args:
        name: Environment name or shortcut. Supports the same shortcuts as
            :func:`parse_env_name`.
        **kwargs: Additional keyword arguments passed to ``gym.make()``.

    Returns:
        Created Gymnasium environment instance.

    Example:
        >>> env = create_env("cart", render_mode="human")
        >>> env = create_env("LunarLander-v2")

    See Also:
        :func:`parse_env_name`: Environment name parsing function
        :func:`gym.make`: Gymnasium environment creation function
    """
    if name in env_short_names:
        return gym.make(env_short_names[name], **kwargs)
    return gym.make(name, **kwargs)


_seed_counter = 0


@deprecated("in favor of callback")
def env_creator_with_seed(config: EnvContext):
    """
    Creates an environment with seed

    Deprecated in favor of callback.
    """
    # NOTE: DO NOT MODIFY CONFIG; reused for VectorEnv
    this_env_config = config.copy()
    seed: int = this_env_config.pop("seed")
    env_type: str = this_env_config.pop("env_type")

    # If using multiple workers, use different seeds for workers with a higher index
    global _seed_counter  # noqa: PLW0603
    if config.num_workers and config.worker_index:
        mixed_seed = np.random.SeedSequence(
            seed,
            spawn_key=(config.worker_index, _seed_counter),  # type: ignore[attr-defined]
        ).generate_state(1)[0]
    else:
        mixed_seed = np.random.SeedSequence(
            seed,
            spawn_key=(0, _seed_counter),
        ).generate_state(1)[0]
    _seed_counter += 1
    _logger.info("Environment seed: %s", mixed_seed)
    geni = np.random.Generator(np.random.PCG64(mixed_seed))

    env = gym.make(env_type, **this_env_config)

    # TODO: Create vector env here

    env.np_random = geni
    _logger.debug(
        "Creating env with seed %s from env_seed=%s for worker idx %s/%s; count=%s.",
        # "Sample obs %s",
        mixed_seed,
        seed,
        config.worker_index,
        config.num_workers,
        _seed_counter,
        # env.reset(),
    )
    return env


register_env("seeded_env", env_creator_with_seed)


def create_env_for_config(config: AlgorithmConfig, env_spec: str | gym.Env):
    """
    Creates an initial environment for the given config.env.

    If it is a `seeded_env` it will create a config from `env_spec` instead.
    """
    if isinstance(config.env, str) and config.env != "seeded_env":
        init_env = gym.make(config.env)
    elif config.env == "seeded_env":
        if isinstance(env_spec, str):
            init_env = gym.make(env_spec)
        else:
            init_env = env_spec
    else:
        assert not TYPE_CHECKING or config.env
        init_env = gym.make(config.env.unwrapped.spec.id)  # pyright: ignore[reportOptionalMemberAccess]
    return init_env



"""NOTE: Experimental, edge cases not tested yet."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Collection

from ray.rllib.algorithms.ppo.torch.ppo_torch_learner import PPOTorchLearner
from ray.rllib.utils.annotations import override
from ray.rllib.utils.typing import ParamDict

if TYPE_CHECKING:
    import torch
    from ray.rllib.utils.typing import (
        ModuleID,
        ParamDict,
        ParamRef,
        TensorType,
    )

_logger = logging.getLogger(__name__)


class PPOTorchLearnerWithGradientAccumulation(PPOTorchLearner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._step_count = 0
        self._gradient_updates = 0
        self._last_gradient_update_step: int | None = None
        self._params: dict[ParamRef, torch.Tensor]  # pyright: ignore[reportIncompatibleVariableOverride]
        # TODO: Test checkpoint loading

    # TODO:
    # [ ] What about last batches if not divisible by accumulate_gradients_every? Reset, keep until next batch?
    # [ ] Add scaling: mean (sum divided by accumulate_gradients_every) or just sum
    @override(PPOTorchLearner)
    def compute_gradients(self, loss_per_module: dict[ModuleID, TensorType], **kwargs) -> ParamDict:  # noqa: ARG002
        self._step_count += 1
        accumulate_gradients_every = self.config.learner_config_dict["accumulate_gradients_every"]
        update_gradients_this_step = self._step_count % accumulate_gradients_every == 0

        if (self._step_count - 1) % accumulate_gradients_every == 0:
            for optim in self._optimizer_parameters:
                # `set_to_none=True` is a faster way to zero out the gradients.
                optim.zero_grad(set_to_none=True)

        if self._grad_scalers is not None:
            total_loss = sum(self._grad_scalers[mid].scale(loss) for mid, loss in loss_per_module.items())
        else:
            total_loss = sum(loss_per_module.values())  # pyright: ignore

        # If we don't have any loss computations, `sum` returns 0. Type will not be a pure int in this case
        if isinstance(total_loss, int):
            assert total_loss == 0
            return {}

        total_loss.backward()
        if update_gradients_this_step:
            self._gradient_updates += 1
            self._last_gradient_update_step = self._step_count
            _logger.debug("Updating gradients for step %s", self._step_count)
            if accumulate_gradients_every != 1:
                grads = {
                    pid: p.grad / accumulate_gradients_every if p.grad is not None else p.grad
                    for pid, p in self._params.items()
                }
            else:  # No accumulation
                grads = {pid: p.grad for pid, p in self._params.items()}
            return grads  # pyright: ignore[reportReturnType]  # contains None
        _logger.debug(
            "Skipping gradient update for step %s, accumulating gradients",
            self._step_count,
        )
        return {}

    def apply_gradients(self, gradients_dict: ParamDict) -> None:
        if self._step_count % self.config.learner_config_dict["accumulate_gradients_every"] == 0:
            # Calls optimizer.step(), scaler.step() and update if applicable
            super().apply_gradients(gradients_dict)

    def after_gradient_based_update(self, *, timesteps: dict[str, Any]) -> None:
        # TODO: Discuss: Scheduler is updated based on non-exact sampled timesteps.
        # if self._step_count % self.config.learner_config_dict["accumulate_gradients_every"] == 0:
        # Updates KL coefficients, and LR Schedulers; if doing this every step batch_size
        # accumulation is not exact to multipling batch size - but should be minor influence if any.
        super().after_gradient_based_update(timesteps=timesteps)

    def get_state(
        self,
        components: str | Collection[str] | None = None,
        *,
        not_components: str | Collection[str] | None = None,
        **kwargs,
    ):
        state_dict = super().get_state(components, not_components=not_components, **kwargs)
        state_dict.update(
            {
                "step_count": self._step_count,
                "gradient_updates": self._gradient_updates,
                "last_gradient_update_step": self._last_gradient_update_step,
            }
        )
        return state_dict

    def set_state(self, state) -> None:
        super().set_state(state)
        self._step_count = state.get("step_count", 0)
        self._gradient_updates = state.get("gradient_updates", 0)
        self._last_gradient_update_step = state.get("last_gradient_update_step", None)



"""High-level utilities for running Ray Tune experiments and local debugging.

This module provides the main entry points for executing Ray RLlib experiments,
both through Ray Tune for distributed hyperparameter optimization and locally
for debugging and testing. It handles the setup, execution, and result processing
of machine learning experiments.

The module bridges experiment setup classes with actual execution, providing
both production-ready distributed training capabilities and development-friendly
local execution modes for rapid iteration and debugging.

Key Components:
    - :func:`run_tune`: Main function for running distributed experiments with Ray Tune
    - Local execution utilities for debugging without Ray Tune overhead
    - Automatic parameter sampling and test mode support
    - Integration with experiment setup classes and trainable functions

This module is typically used as the final step in experiment configuration,
after setting up algorithms, hyperparameters, and training configurations.
"""

from __future__ import annotations

import logging
import sys
import time
from inspect import isclass
from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar

from ray.rllib.algorithms.algorithm import Algorithm
from ray.rllib.algorithms.algorithm_config import AlgorithmConfig

from ray_utilities.config import DefaultArgumentParser
from ray_utilities.misc import raise_tune_errors
from ray_utilities.random import seed_everything
from ray_utilities.training.default_class import TrainableBase

if sys.version_info < (3, 11):
    from exceptiongroup import BaseExceptionGroup

if TYPE_CHECKING:
    from ray.rllib.algorithms import Algorithm, AlgorithmConfig
    from ray.tune.result_grid import ResultGrid

    from ray_utilities.config import DefaultArgumentParser
    from ray_utilities.setup import ExperimentSetupBase
    from ray_utilities.training.default_class import TrainableBase
    from ray_utilities.typing import TestModeCallable
    from ray_utilities.typing.trainable_return import TrainableReturnData

logger = logging.getLogger(__name__)

_SetupT = TypeVar("_SetupT", bound="ExperimentSetupBase[DefaultArgumentParser, AlgorithmConfig, Algorithm]")
"""A TypeVar for ExperimentSetupBase with specific type parameters."""


def _run_without_tuner(
    setup: _SetupT,
    trainable: type[TrainableBase[Any, Any, Any]] | Callable[[dict], TrainableReturnData],
    test_mode_func: Optional[TestModeCallable[_SetupT]] = None,
) -> TrainableReturnData:
    """Test and debug mode function that does not run a Tuner instance but locally."""
    # will spew some warnings about train.report
    func_name = getattr(test_mode_func, "__name__", repr(test_mode_func)) if test_mode_func else trainable.__name__
    print(f"-- FULL TEST MODE running {func_name} --")
    logger.info("-- FULL TEST MODE --")
    import ray.tune.search.sample  # noqa: PLC0415 # import lazy

    # Sample the parameters when not entering via tune
    params = {
        k: v.sample() if isinstance(v, ray.tune.search.sample.Domain) else v for k, v in setup.param_space.items()
    }
    setup.param_space.update(params)
    # Possibly set RAY_DEBUG=legacy
    if isclass(trainable):
        # If trainable is a class, instantiate it with the sampled parameters
        trainable_instance = trainable(setup.sample_params())
        logger.warning(
            "[TESTING] Using a Trainable class, without a Tuner relying on a stopper or 'done' return value."
        )

        tuner = setup.create_tuner()
        assert tuner._local_tuner
        stopper = tuner._local_tuner.get_run_config().stop
        while True:
            result = trainable_instance.train()
            if callable(stopper):
                # If stop is a callable, call it with the result
                if stopper("NA", result):  # pyright: ignore[reportArgumentType]
                    break
            # If stop is not a callable, check if it is reached
            elif result.get("done", False):
                break
        from ray_utilities.postprocessing import create_log_metrics  # noqa: PLC0415, circular import

        if "config" in result:
            result["config"].setdefault("_train_batch_size_per_learner", setup.config.train_batch_size_per_learner)
        return create_log_metrics(result, log_stats=setup.args.log_stats)
    if test_mode_func:
        return test_mode_func(trainable, setup)
    return trainable(setup.sample_params())


def run_tune(
    setup: _SetupT | type[_SetupT],
    test_mode_func: Optional[TestModeCallable[_SetupT]] = None,
    *,
    raise_errors: bool = True,
) -> TrainableReturnData | ResultGrid:
    """Execute hyperparameter tuning for Ray RLlib experiments.

    This function orchestrates the complete tuning workflow, including experiment setup,
    test mode handling, seed configuration, and result management. It provides the main
    entry point for running hyperparameter optimization experiments.

    Args:
        setup: Experiment setup instance or class. If a class is provided, it will be
            instantiated with default parameters. Must inherit from :class:`ExperimentSetupBase`.
        test_mode_func: Optional function to execute in test mode instead of full tuning.
            Should accept the trainable and setup as arguments and return results.
        raise_errors: Whether to raise exceptions encountered during tuning. If False,
            errors are logged but execution continues.

    Returns:
        Ray Tune results from the hyperparameter optimization process, or test results
        if running in test mode. Returns a :class:`ray.tune.ResultGrid` for full tuning
        or :class:`TrainableReturnData` for test mode execution.

    Examples:
        Basic usage with a setup class:

        >>> from ray_utilities.setup import PPOSetup
        >>> results = run_tune(PPOSetup)

        With a configured setup instance:

        >>> setup = PPOSetup()
        >>> setup.config.lr = 0.001
        >>> results = run_tune(setup)

        Using test mode:

        >>> def test_func(trainable, setup):
        ...     return trainable(setup.sample_params())
        >>> results = run_tune(PPOSetup, test_mode_func=test_func)

    Note:
        - Test mode is activated when ``args.test=True`` and ``args.not_parallel=True``
        - Offline experiment tracking uploads are handled automatically after completion.
          When ``args.comet`` or ``args.wandb`` are set to ``offline+upload``.
        - Seeds are automatically configured across Ray, torch, and numpy when specified

    See Also:
        :class:`ray_utilities.setup.ExperimentSetupBase`: Base class for experiment setups
        :func:`_run_without_tuner`: Test mode execution function
        :func:`ray.tune.run`: Underlying Ray Tune execution
    """
    # full parser example see: https://github.com/ray-project/ray/blob/master/rllib/utils/test_utils.py#L61

    if isinstance(setup, type):
        setup = setup()
    args = setup.get_args()
    if args.seed is not None:
        logger.debug("Setting seed to %s", args.seed)
        _next_seed = seed_everything(env=None, seed=args.seed, torch_manual=True, torch_deterministic=True)
        if setup.config.seed != args.seed:
            with setup.open_config():  # config is frozen
                setup.config.seed = args.seed
    trainable = setup.trainable or setup.create_trainable()

    # -- Test --
    if args.test and args.not_parallel:
        return _run_without_tuner(setup=setup, trainable=trainable, test_mode_func=test_mode_func)
    # Use tune.with_parameters to pass large objects to the trainable

    results = None
    fit_error = None
    fallback_error = None
    tuner = setup.create_tuner()
    try:
        results = tuner.fit()
    except (KeyboardInterrupt, Exception) as e:
        if isinstance(e, KeyboardInterrupt):
            # In case we do not want to upload results
            logger.info(
                "Tuning interrupted. Will try to upload gathered results now. "
                "Waiting 2 sec before continuing... "
                "Press Ctrl + C again to exit immediately."
            )
            time.sleep(2)
        fit_error = e
        logger.error("Error occurred during tuning: %s", fit_error)
        try:
            results = tuner.get_results()
        except (RuntimeError, Exception) as e2:
            logger.error("Error occurred while getting results: %s", e)
            fallback_error = e2
    if not results:
        logger.warning("No results returned from the tuner.")
    setup.upload_offline_experiments(results, tuner)
    if raise_errors and results is not None:
        raise_tune_errors(results)
    if fallback_error and fit_error:
        # Use Base in case of KeyboardInterrupt
        raise BaseExceptionGroup(
            f"Encountered an error {fit_error} and could not call get_results {fallback_error!r}",
            [fit_error, fallback_error],
        )
    if fit_error:
        raise fit_error
    assert results is not None

    return results


if __name__ == "__main__":
    # For testing purposes, run the function with a dummy setup
    from ray_utilities.setup import PPOSetup

    dummy_setup = PPOSetup()
    run_tune(dummy_setup)



from __future__ import annotations

from dataclasses import field
from typing import TYPE_CHECKING

import jax.numpy as jnp
from chex import dataclass
from flax.training.train_state import TrainState as FlaxTrainState

if TYPE_CHECKING:
    import chex
    import jax


class ExtendedTrainState(FlaxTrainState):
    grad_accum: jnp.ndarray


@dataclass(kw_only=True, frozen=True, eq=False)
class Indices:
    """Frozen and hashable variant of the indices dict used by SYMPOL to allow using them as static_args"""

    features_by_estimator: jax.Array = field(hash=False, compare=True)
    # do not use repr because long
    path_identifier_list: jax.Array = field(hash=False, compare=True, repr=False)
    internal_node_index_list: jax.Array = field(hash=False, compare=True, repr=False)

    def __hash__(self) -> int:
        if self._hash is None:  # type: ignore
            object.__setattr__(
                self,
                "_hash",
                hash(
                    (
                        tuple(a.item() for entry in self.features_by_estimator for a in entry),
                        tuple(a.item() for entry in self.path_identifier_list for a in entry),
                        tuple(a.item() for entry in self.internal_node_index_list for a in entry),
                    )
                ),
            )
        return self._hash

    def __post_init__(self) -> None:
        object.__setattr__(self, "_hash", None)
        if TYPE_CHECKING:
            self._hash: int = None  # pyright: ignore
        self.__hash__()

    # backwards compatibility
    def __getitem__(self, key: str) -> chex.Array:
        return getattr(self, key)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Indices):
            return NotImplemented
        out = (
            jnp.array_equal(self.features_by_estimator, other.features_by_estimator).item()
            and jnp.array_equal(self.path_identifier_list, other.path_identifier_list).item()
            and jnp.array_equal(self.internal_node_index_list, other.internal_node_index_list).item()
        )
        return out



"""Advanced Population Based Training scheduler for Ray Tune hyperparameter optimization.

This module provides :class:`ReTuneScheduler`, an enhanced version of Ray Tune's
PopulationBasedTraining that supports grid search mutations, flexible quantile
fractions, and improved trial management for reinforcement learning experiments.

Key Components:
    - :class:`ReTuneScheduler`: Enhanced PBT scheduler with advanced features
    - :func:`_grid_search_sample_function`: Grid search sampling utilities
    - Integration with Ray Tune's trial and search algorithm framework

The scheduler extends the standard PBT approach with support for deterministic
grid search mutations and more flexible population management strategies.
"""

from __future__ import annotations

import logging
import math
import random
from functools import partial
from itertools import cycle
from typing import TYPE_CHECKING, Any, Callable, Container, Dict, Generic, List, Optional, Tuple, TypeVar, cast

from ray.train._internal.session import _FutureTrainingResult, _TrainingResult
from ray.tune.experiment import Trial
from ray.tune.result import TRAINING_ITERATION  # pyright: ignore[reportPrivateImportUsage]
from ray.tune.schedulers.pbt import PopulationBasedTraining

from ray_utilities.constants import FORK_FROM, PERTURBED_HPARAMS
from ray_utilities.misc import build_nested_dict, flatten_mapping_with_path, get_value_by_path
from ray_utilities.typing import ForkFromData, Forktime

if TYPE_CHECKING:
    from collections.abc import Iterable, MutableMapping

    from ray.tune.execution.tune_controller import TuneController
    from ray.tune.schedulers.pbt import _PBTTrialState
    from ray.tune.search.sample import Domain

    from ray_utilities.typing.algorithm_return import AlgorithmReturnData
    from ray_utilities.typing.metrics import FlatLogMetricsDict


logger = logging.getLogger(__name__)
_T = TypeVar("_T")


MAX_SKIP_LIST_LENGTH = 10000


if TYPE_CHECKING:
    from typing import type_check_only

    @type_check_only
    class _PBTTrialState2(_PBTTrialState):
        def __init__(self, trial: Trial):
            super().__init__(trial)
            self.last_training_iteration: int  # not present before on_trial_add
            """The training iteration at which the last result was reported."""


def _grid_search_sample_function(grid_search_space: Iterable[_T], *, repeat=True) -> Callable[[], _T]:
    """Create a function for sampling from a grid search space.

    Returns a parameterless function that yields grid search values either cyclically
    (with repetition) or once through the space (without repetition).

    Args:
        grid_search_space: Iterable containing the values to sample from.
        repeat: If True, cycle through values infinitely. If False, each value
            is returned once and then StopIteration is raised.

    Returns:
        A parameterless function that returns the next grid search sample.
        When repeat=False, the function raises StopIteration after all values
        have been returned once.

    Example:
        >>> sampler = _grid_search_sample_function([1, 2, 3], repeat=True)
        >>> sampler()  # Returns 1
        >>> sampler()  # Returns 2
        >>> sampler()  # Returns 3
        >>> sampler()  # Returns 1 (cycles back)
    """
    if repeat:
        cycler = cycle(grid_search_space)

        def cyclic_grid_iterator():
            return next(cycler)

        return cyclic_grid_iterator
    grid_search_space = list(grid_search_space)

    def grid_iterator():
        try:
            return grid_search_space.pop(0)
        except IndexError as e:
            raise StopIteration from e

    return grid_iterator


def _debug_dump_new_config(new_config: dict, mutation_keys: list[str]):
    logger.info("New config after perturbation %s", new_config)
    new_config[PERTURBED_HPARAMS] = {k: new_config[k] for k in mutation_keys}
    return new_config


class CyclicMutation(Generic[_T]):
    def __init__(self, values: Iterable[_T], skip: Optional[Container[_T]] = None, *, disable_skip: bool = False):
        self._cycler = cycle(values)
        self._values = list(values)
        self.skip = skip
        self.disable_skip = disable_skip
        self._warn_possible_infinite_loop()

    def update_values(self, new_values: Iterable[_T]):
        self._values = list(new_values)
        self._cycler = cycle(self._values)
        self._warn_possible_infinite_loop()

    def update_skip(self, new_values: Container[_T] | None):
        self.skip = new_values
        self._warn_possible_infinite_loop()

    def _warn_possible_infinite_loop(self) -> bool | None:
        try:
            if (
                self.skip and not self.disable_skip and len(self.skip) >= len(self._values)  # pyright: ignore[reportArgumentType]
            ):
                logger.warning(
                    "CyclicMutation skip list %s has length >= values %s. "
                    "This may cause an infinite loop when sampling.",
                    self.skip,
                    self._values,
                )
                return True
        except TypeError:
            # self.skip is only a container, not sized
            return None
        return False

    def __call__(self) -> _T:
        v = next(self._cycler)
        if self.disable_skip or self.skip is None:
            return v
        i = 0
        while self.skip and v in self.skip:
            v = next(self._cycler)
            i += 1
            if i > MAX_SKIP_LIST_LENGTH:
                loop = self._warn_possible_infinite_loop()
                if loop or loop is None:
                    raise RuntimeError(
                        "CyclicMutation appears to be stuck in an infinite loop due to skip list. "
                        "Increase MAX_SKIP_LIST_LENGTH."
                    )
            elif i % 1000 == 0:
                loop = self._warn_possible_infinite_loop()
                logger.warning(
                    "CyclicMutation still searching for non-skipped value after %s attempts. Is infinite loop: %s",
                    i,
                    loop or ("unknown (not Sized)" if loop is None else "no"),
                )
        return v


class TopPBTTrialScheduler(PopulationBasedTraining):
    """Enhanced Population Based Training scheduler with grid search and flexible quantiles.

    This scheduler extends Ray Tune's PopulationBasedTraining with support for grid search
    mutations, and improved trial management for reinforcement
    learning experiments.
    The most prominent change is that all trials outside of the top quantile are exploited.
    That is is changes, for a quantile fraction of 0.1, all 90% of trials are exploited,
    instead of only the lowest 10% and keeping the other 80% as is.

    Key enhancements over standard PBT:
        - Grid search mutations for deterministic hyperparameter exploration
        - Custom exploration functions with mutation tracking
        - Enhanced logging and debugging capabilities

    The scheduler maintains compatibility with the standard PBT interface while providing
    additional flexibility for advanced hyperparameter optimization strategies.

    Args:
        time_attr: Attribute to use for time progression tracking.
        metric: Metric name to optimize (e.g., "episode_reward_mean").
        mode: Optimization mode, either "max" or "min".
        perturbation_interval: Number of time units between perturbations.
        burn_in_period: Time units before perturbations begin.
        hyperparam_mutations: Dictionary mapping hyperparameter names to mutation
            specifications. Supports grid_search definitions for deterministic sampling.
        quantile_fraction: Fraction of population to consider as "top performers".
        resample_probability: Probability of resampling parameters during perturbation.
        perturbation_factors: Tuple of (lower, upper) factors for parameter perturbation.
        custom_explore_fn: Optional custom function for exploration logic.
        log_config: Whether to log configuration changes.
        require_attrs: Whether to require time_attr and metric in results.
        synch: Whether to use synchronous perturbation.

    Example:
        >>> scheduler = ReTuneScheduler(
        ...     metric="episode_reward_mean",
        ...     mode="max",
        ...     perturbation_interval=50000,
        ...     hyperparam_mutations={
        ...         "lr": {"grid_search": [1e-4, 5e-4, 1e-3]},
        ...         "batch_size": {"grid_search": [64, 128, 256]},
        ...     },
        ...     quantile_fraction=0.1,  # Keep and exploit the top 10% of trials
        ... )

    Note:
        Grid search spaces in hyperparam_mutations are automatically converted to
        sampling functions that cycle through the provided values.

    See Also:
        :class:`ray.tune.schedulers.pbt.PopulationBasedTraining`: Base PBT scheduler
        :func:`_grid_search_sample_function`: Grid search sampling utilities
    """

    additional_config_keys = (
        FORK_FROM,
        "_top_pbt_is_in_upper_quantile",
        "_top_pbt_perturbed",
    )
    """Keys inserted into the config of trials to track PBT state."""

    def __init__(
        self,
        *,
        time_attr: str = "current_step",
        metric: str | None = None,
        mode: str | None = "max",
        perturbation_interval: float = 100_000,
        burn_in_period: float = 0,
        hyperparam_mutations: Optional[
            MutableMapping[str, dict[str, Any] | list | tuple | Callable[..., Any] | Domain]
        ] = None,
        # Use only very best trial # TODO: Should probably use more but double trials.
        quantile_fraction: float = 0.1,  # 0.25,  # 0 for no exploit -> no top trials, 0.99 for only exploit top trial
        resample_probability: float = 1.0,  # Always resample, assume grid_search in hyperparam_mutations # TODO: alt use custom_explore_fn with new value as input
        perturbation_factors: Tuple[float, float] = (0.8, 1.2),
        custom_explore_fn: Callable[..., Any] | None = None,
        log_config: bool = True,
        require_attrs: bool = True,
        synch: bool = False,
    ):
        if custom_explore_fn is None:  # Otherwise use a wrapper
            # XXX
            if hyperparam_mutations:
                custom_explore_fn = partial(_debug_dump_new_config, mutation_keys=list(hyperparam_mutations.keys()))
        # TODO: Do we want to reload the env_seed?
        self._trial_state: dict[Trial, _PBTTrialState2]  # pyright: ignore[reportIncompatibleVariableOverride]
        if hyperparam_mutations:  # either hyperparam_mutations or custom_explore_fn must be passed
            for k, v in hyperparam_mutations.items():
                if isinstance(v, dict) and "grid_search" in v:
                    hyperparam_mutations[k] = _grid_search_sample_function(v["grid_search"])
        super().__init__(
            time_attr,
            metric,
            mode,
            perturbation_interval,
            burn_in_period,
            hyperparam_mutations,  # pyright: ignore[reportArgumentType]
            quantile_fraction,
            resample_probability,
            perturbation_factors,
            custom_explore_fn,  # only used on explore (see _exploit function, get_new_config)
            log_config,
            require_attrs,
            synch,
        )
        # Store assignments for exploit distribution
        self._exploit_assignments = {}
        self._current_assignments: dict[Trial, Trial] | None = None

    def on_trial_add(self, tune_controller: TuneController, trial: Trial):
        """Updates the trials config with hyperparam_mutations"""
        # Adds a new trial with config updated based on hyperparam_mutations
        super().on_trial_add(tune_controller, trial)
        if FORK_FROM in trial.config:
            self._trial_state[trial].last_training_iteration = trial.config[FORK_FROM].get("fork_training_iteration", 0)
        else:
            self._trial_state[trial].last_training_iteration = 0

    def _quantiles(self) -> Tuple[List[Trial], List[Trial]]:
        """Returns trials in the lower and upper `quantile` of the population.

        If there is not enough data to compute this, returns empty lists.

        This method allows for quantile_fraction > 0.5 as well.
        All trials outside the top quantile are considered in the lower quantile,
        meaning they will exploit the top-performing trials.
        """
        trials: list[Trial] = []
        for trial, state in self._trial_state.items():
            logger.debug("Trial %s, state %s", trial, state)
            if trial.is_finished():
                logger.debug("Trial %s is finished", trial)
            if state.last_score is not None and not math.isnan(state.last_score) and not trial.is_finished():
                trials.append(trial)

        # Sort trials by score; _save_trial_state takes care of mode (multiply by -1 if min mode)
        trials.sort(key=lambda t: self._trial_state[t].last_score)

        if len(trials) <= 1:
            return [], []

        # Calculate number of trials in top quantile
        num_top_trials = max(1, math.ceil(len(trials) * self._quantile_fraction))

        if num_top_trials > len(trials) / 2:
            num_top_trials = math.floor(len(trials) / 2)
        top_trials = trials[-num_top_trials:]
        # all other trials will exploit top trials
        bottom_trials = trials[:-num_top_trials]

        logger.debug("Split trials: %s in top quantile, %s in bottom quantile", len(top_trials), len(bottom_trials))

        return bottom_trials, top_trials

    def _distribute_exploitation(self, lower_quantile: List[Trial], upper_quantile: List[Trial]) -> Dict[Trial, Trial]:
        """Distribute the exploitation of top trials evenly among bottom trials.

        Args:
            lower_quantile: List of trials that will exploit top trials
            upper_quantile: List of top-performing trials to be exploited

        Returns:
            Dictionary mapping each bottom trial to the top trial it should exploit
        """
        if not upper_quantile or not lower_quantile:
            return {}

        assignments: dict[Trial, Trial] = {}
        # Create cyclic assignment to ensure even distribution
        top_trials_cycle = cycle(upper_quantile)

        for trial in lower_quantile:
            assignments[trial] = next(top_trials_cycle)

        # Log the distribution
        distribution = {trial.trial_id: 1 for trial in upper_quantile}
        for top in assignments.values():
            distribution[top.trial_id] += 1

        logger.debug("Exploitation distribution: %s", distribution)
        return assignments

    @classmethod
    def _deep_update_cyclic_mutation(
        cls,
        mutations: dict[str, CyclicMutation[_T] | dict[str, Any] | Any],
        new_skip: Optional[dict[str, Container[_T] | dict[str, Any] | None]],
    ) -> None:
        if new_skip is None:
            return
        for key, distribution in mutations.items():
            if isinstance(distribution, dict) and key in new_skip:
                deeper_skip = new_skip[key]
                assert isinstance(deeper_skip, dict)
                cls._deep_update_cyclic_mutation(distribution, cast("dict[str, Any]", deeper_skip))
            elif isinstance(distribution, CyclicMutation):
                deeper_skip = new_skip[key]
                assert not isinstance(deeper_skip, dict)
                cast("CyclicMutation[_T]", distribution).update_skip(deeper_skip)

    def _get_current_best_mutations(self, upper_quantile: list[Trial]):
        """
        Get the values of hyperparameter mutations for the best trials.

        Use to update the skip lists in CyclicMutation instances.
        """
        # Get all paths to leaves in the mutation space

        flat_mutation_keys = flatten_mapping_with_path(self._hyperparam_mutations)
        mutation_paths = [path for path, _ in flat_mutation_keys]

        # Collect values for each path from all trial configs
        path_to_values: dict[tuple, list] = {path: [] for path in mutation_paths}
        for trial in upper_quantile:
            config = trial.config
            for path in mutation_paths:
                try:
                    value = get_value_by_path(config, path)
                except KeyError:
                    value = None
                path_to_values[path].append(value)

        return build_nested_dict(mutation_paths, path_to_values)

    def _checkpoint_or_exploit(
        self,
        trial: Trial,
        tune_controller: "TuneController",
        upper_quantile: List[Trial],
        lower_quantile: List[Trial],
    ):
        """Checkpoint if in upper quantile, exploits if in lower.

        For trials in the lower quantile, evenly distribute which top trial
        they exploit to ensure balanced exploitation.
        """
        state = self._trial_state[trial]
        # Remove any fork controlling keys from the config
        for k in self.additional_config_keys:
            trial.config.pop(k, None)

        # Create exploitation assignments if needed
        self._current_assignments = self._distribute_exploitation(lower_quantile, upper_quantile)
        # Update any CyclicMutation skip lists based on current top trials, to not resample these values.
        new_skips = self._get_current_best_mutations(upper_quantile)
        logger.debug("Updating CyclicMutation skip lists to %s", new_skips)
        self._deep_update_cyclic_mutation(self._hyperparam_mutations, new_skip=new_skips)
        # TODO: a trial should just use the batch size assigned that it already has, i.e. actually no mutation, just load best checkpoint
        # FIXME: next perturbation interval does not increase linearly. Trials train longer and longer as well

        # Create a checkpoint for all trials
        logger.debug("Instructing %s to save.", trial)
        checkpoint = tune_controller._schedule_trial_save(trial, result=state.last_result)

        # Only keep checkpoints for top trials
        if trial in upper_quantile:
            # TODO: check this again in TuneControl the trial last result is only updated after the scheduler
            # callback. So, we override with the current result.
            logger.debug("Trial %s is in upper quantile. Saving checkpoint.", trial)
            if trial.status == Trial.PAUSED:
                if trial.temporary_state.saving_to and isinstance(
                    trial.temporary_state.saving_to, _FutureTrainingResult
                ):
                    logger.debug("Trial %s is still saving.", trial)
                    state.last_checkpoint = trial.temporary_state.saving_to
                else:
                    # Paused trial will always have an in-memory checkpoint.
                    logger.debug("Trial %s is paused. Use last available checkpoint %s.", trial, trial.checkpoint)
                    state.last_checkpoint = trial.checkpoint
            else:
                logger.debug("Keeping checkpoint of trial %s for exploit.", trial)
                # TODO: # FIXME does this create two checkpoint with Trainable Auto saving?
                # state.last_checkpoint = tune_controller._schedule_trial_save(trial, result=state.last_result)
                state.last_checkpoint = checkpoint

            self._num_checkpoints += 1
            trial.config["_top_pbt_is_in_upper_quantile"] = True
        else:
            state.last_checkpoint = None  # not a top trial

        # If in lower quantile, exploit a top trial based on our distribution
        if trial in lower_quantile:
            # Get the assigned top trial to exploit
            trial_to_clone = self._current_assignments.get(trial)

            if not trial_to_clone:
                logger.warning("No exploitation assignment for trial %s. Using random selection.", trial)
                trial_to_clone = random.choice(upper_quantile)

            assert trial is not trial_to_clone
            assert trial_to_clone in upper_quantile
            clone_state = self._trial_state[trial_to_clone]
            last_checkpoint = clone_state.last_checkpoint

            logger.debug(
                "Trial %s is exploiting trial %s (rank %s/%s).",
                trial,
                trial_to_clone,
                upper_quantile.index(trial_to_clone) + 1,
                len(upper_quantile),
            )

            if isinstance(last_checkpoint, _FutureTrainingResult):
                training_result: _TrainingResult | None = last_checkpoint.resolve()

                if training_result:
                    clone_state.last_result = training_result.metrics
                    clone_state.last_checkpoint = training_result.checkpoint
                    last_checkpoint = clone_state.last_checkpoint
                else:
                    logger.error(
                        "PBT-scheduled checkpoint save resolved to None. Trial "
                        "%s didn't save any checkpoint before "
                        "and can't be exploited.",
                        trial_to_clone,
                    )
                    last_checkpoint = None

            if not last_checkpoint:
                logger.info("[pbt]: no checkpoint for trial %s. Skip exploit for Trial %s", trial_to_clone, trial)
                return
            self._exploit(tune_controller, trial, trial_to_clone)
            # Mark trial as perturbed
            for k in self.additional_config_keys:
                trial.config.pop(k, None)
            # Set info which trial was forked from
            fork_data: ForkFromData = {
                "parent_id": trial_to_clone.trial_id,
                "parent_trial": trial_to_clone,
                "parent_training_iteration": self._trial_state[trial_to_clone].last_training_iteration,  # pyright: ignore[reportAttributeAccessIssue]
                "parent_time": Forktime(self._time_attr, self._trial_state[trial_to_clone].last_train_time),
                "controller": self.__class__.__name__,
            }
            trial.config[FORK_FROM] = fork_data
            trial.invalidate_json_state()
        else:
            for k in self.additional_config_keys:
                trial.config.pop(k, None)
            trial.config["_top_pbt_perturbed"] = False

    def _save_trial_state(
        self, state: _PBTTrialState | _PBTTrialState2, time: int, result: AlgorithmReturnData | dict, trial: Trial
    ):
        score = super()._save_trial_state(state, time, cast("dict", result), trial)
        # Save training iteration for the step for loggers like WandB / Comet
        state.last_training_iteration = result[TRAINING_ITERATION]  # pyright: ignore[reportAttributeAccessIssue]
        return score

    def reset_exploitations(self):
        """Reset the current exploitation assignments.

        This should be called at the beginning of each perturbation round.
        """
        self._current_assignments = None

    def on_trial_complete(
        self, tune_controller: TuneController, trial: Trial, result: FlatLogMetricsDict | dict[str, Any]
    ):
        """Handle completed trial by cleaning up assignments."""
        # Happens on trial.Terminated, e.g. stopper returned True
        super().on_trial_complete(
            tune_controller,
            trial,
            cast("dict[str, object]", result),
        )
        # Reset assignments when trials complete to ensure proper redistribution
        self.reset_exploitations()



# ruff: noqa: ARG002

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Optional

from ray.rllib.connectors.connector_v2 import ConnectorV2

if TYPE_CHECKING:
    from ray.rllib.core.learner.learner import Learner
    from ray.rllib.core.rl_module.rl_module import RLModule
    from ray.rllib.utils.metrics.metrics_logger import MetricsLogger
    from ray.rllib.utils.typing import EpisodeType


class DebugConnector(ConnectorV2):
    """Debug connector for Ray RLlib pipeline inspection and troubleshooting.

    This connector provides debugging capabilities for Ray RLlib's data processing
    pipeline by adding logging, inspection points, and optional breakpoints at
    specific stages of the training workflow. It can be inserted anywhere in the
    connector pipeline to monitor data flow and help with troubleshooting.

    The connector logs detailed information about the data passing through it,
    including batches, episodes, exploration settings, and metrics. It supports
    both debug-level logging for continuous monitoring and warning-level logging
    with detailed dumps for specific debugging sessions.

    Features:
        - Configurable logging levels and message detail
        - Optional breakpoint insertion for interactive debugging
        - Non-intrusive data flow monitoring
        - Named connector instances for pipeline identification

    Args:
        name: Identifier for this connector instance, used in log messages
            to distinguish between multiple debug connectors in the pipeline.
        *args: Additional positional arguments passed to parent :class:`ConnectorV2`.
        **kwargs: Additional keyword arguments passed to parent :class:`ConnectorV2`.

    Example:
        >>> # Add to environment-to-module connector pipeline
        >>> config.env_to_module_connector = lambda obs_space, act_space: [
        ...     DebugConnector(name="env_to_module_debug"),
        ...     # ... other connectors
        ... ]

    Note:
        The debug logging and breakpoint functionality can be controlled by
        modifying the conditional checks in the ``__call__`` method. Set the
        condition to ``True`` to enable detailed logging dumps.

    See Also:
        :class:`ray.rllib.connectors.connector_v2.ConnectorV2`: Base connector class
        :class:`ray.rllib.core.rl_module.rl_module.RLModule`: RL module interface
    """

    def __init__(self, *args, name: str, **kwargs):
        super().__init__(*args, **kwargs)
        self._logger = logging.getLogger(__name__)
        self._name = name
        self._logger.setLevel(logging.DEBUG)

    def __call__(
        self,
        *,
        rl_module: RLModule,
        batch: dict[str, Any],
        episodes: list[EpisodeType],
        explore: Optional[bool] = None,
        shared_data: Optional[dict] = None,
        metrics: Optional[MetricsLogger] = None,
        **kwargs,
    ) -> Any:
        if False:
            self._logger.warning(
                "DebugConnector (%s): \nbatch=%s, \nepisodes=%s, \nexplore=%s, \nshared_data=%s, \nmetrics=%s\n-------",
                self._name,
                batch,
                episodes,
                explore,
                shared_data,
                metrics,
            )
        self._logger.debug("DebugConnector called with batch: %s", batch)
        breakpoint()  # noqa: T100
        return batch


def add_debug_connectors(learner: Learner) -> None:
    """
    Adds DebugConnector to the learner connector pipeline if the config
    has "_debug_connectors" set to True.

    Will clean existing debug connectors (start, end only) if they exist.
    """
    if learner.config.learner_config_dict.get("_debug_connectors", False):
        if not learner._learner_connector:
            # create a learner connector
            learner._learner_connector = learner.config.build_learner_connector(
                input_observation_space=None,
                input_action_space=None,
                device=learner._device,
            )
            learner._learner_connector.append(DebugConnector(name="Learner debug End"))
            return
        # make sure to clean existing debug connectors (start, end only)
        if learner._learner_connector:
            if isinstance(learner._learner_connector.connectors[0], DebugConnector):
                learner._learner_connector.connectors.pop(0)
            if isinstance(learner._learner_connector.connectors[-1], DebugConnector):
                learner._learner_connector.connectors.pop(-1)
        learner._learner_connector.prepend(DebugConnector(name="Learner debug Start"))
        learner._learner_connector.append(DebugConnector(name="Learner debug End"))


if TYPE_CHECKING:
    DebugConnector(name="abc")



from __future__ import annotations

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Iterable, Optional, Sequence, Tuple, Union, cast

import distrax
import jax
import jax.numpy as jnp
import numpy as np
import tree
from distrax import Distribution as DistraxDistribution
from ray.rllib.core.distribution.distribution import Distribution as RllibDistribution
from tensorflow_probability.substrates import jax as tfp
from typing_extensions import Self

if TYPE_CHECKING:
    import chex
    import gymnasium as gym
    from ray.rllib.utils.typing import TensorType

    IntLike = Union[int, np.int16, np.int32, np.int64]
    from tensorflow_probability.substrates.jax import distributions as __tfd

logger = logging.getLogger(__name__)


def _get_tfp_distributions():
    """Type hint support for return value"""
    if TYPE_CHECKING:
        return __tfd
    return tfp.distributions


class RLlibToJaxDistribution(RllibDistribution, distrax.Distribution):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dist = self._get_jax_distribution(*args, **kwargs)

    @abstractmethod
    def _get_jax_distribution(self, *args, **kwargs) -> distrax.Distribution: ...

    def sample(
        self,
        *,
        seed: IntLike | chex.PRNGKey,  # XXX <-- not RLlib interface
        sample_shape: Union[IntLike, Sequence[IntLike]] = (),
        return_logp: bool = False,
        **kwargs,
    ) -> Union[TensorType, tuple[TensorType, TensorType]]:
        if not return_logp:
            return self._dist.sample(seed=seed, sample_shape=sample_shape, **kwargs)
        return self._dist.sample_and_log_prob(seed=seed, sample_shape=sample_shape, **kwargs)

    def logp(self, value: TensorType, **kwargs) -> TensorType:
        return self._dist.log_prob(value, **kwargs)

    def kl(self, other: "RLlibToJaxDistribution | distrax.DistributionLike", **kwargs) -> TensorType:
        if isinstance(other, RLlibToJaxDistribution):
            other = other._dist
        return self._dist.kl_divergence(other, **kwargs)

    @abstractmethod
    def to_deterministic(self) -> Deterministic: ...


class SupportsLogitsMixin(RllibDistribution):
    @classmethod
    def from_logits(cls, logits: TensorType, **kwargs):
        return cls(logits=logits, **kwargs)  # type: ignore[call-arg]


class _EntropyArgCorrector(RllibDistribution):
    def entropy(self, **kwargs) -> TensorType:
        return cast(
            "DistraxDistribution",
            self._dist,  # type: ignore[attr-defined]
        ).entropy(**kwargs)


class Normal(_EntropyArgCorrector, distrax.Normal, RLlibToJaxDistribution):
    def __init__(self, loc: chex.Numeric, scale: chex.Numeric):
        super().__init__(loc, scale)
        self._dist: distrax.Normal

    def _get_jax_distribution(self, *args, **kwargs) -> distrax.Distribution:
        return distrax.Normal(*args, **kwargs)

    def rsample(
        self,
        *,
        seed: IntLike | chex.PRNGKey,  # XXX <-- not RLlib interface
        sample_shape: Tuple[int, ...] = (),
        return_logp: bool = False,
        **kwargs,
    ) -> Union[TensorType, Tuple[TensorType, TensorType]]:
        eps = jax.random.normal(seed, sample_shape, **kwargs)
        if not return_logp:
            return self._dist.loc + eps * self._dist.scale
        # possibly use _sample_from_std_normal
        raise NotImplementedError("rsample with return_logp not implemented")

    @staticmethod
    def required_input_dim(space: gym.Space | gym.spaces.Box, **kwargs) -> int:  # noqa: ARG004
        assert space.shape is not None
        return int(np.prod(space.shape, dtype=np.int32))

    def to_deterministic(self) -> Deterministic:
        return Deterministic(loc=self.loc)


if TYPE_CHECKING:
    Normal(1, 0)


class Categorical(_EntropyArgCorrector, SupportsLogitsMixin, RLlibToJaxDistribution, distrax.Categorical):
    def __init__(
        self,
        probs: Optional[chex.Array] = None,
        logits: Optional[chex.Array] = None,
        dtype: Union[jnp.dtype, type[Any]] = int,
    ):
        # Problem super call call to distrax.Categorical does not forward to RLlibDistribution
        super().__init__(probs=probs, logits=logits, dtype=dtype)
        self.one_hot = _get_tfp_distributions().OneHotCategorical(logits=logits, probs=probs)

    def _get_jax_distribution(
        self,
        probs: Optional[chex.Array] = None,
        logits: Optional[chex.Array] = None,
        dtype: Union[jnp.dtype, type[Any]] = int,
    ) -> distrax.Distribution:
        self._dist: distrax.Categorical
        return distrax.Categorical(probs=probs, logits=logits, dtype=dtype)

    def rsample(self, *, sample_shape=(), return_logp: bool = False, seed=None, **kwargs):
        if return_logp:
            one_hot_sample, logp = self.one_hot.experimental_sample_and_log_prob(sample_shape, seed=seed, **kwargs)
            return jax.lax.stop_gradient(one_hot_sample - self.probs) + self.probs, logp
        one_hot_sample = self.one_hot.sample(sample_shape, seed=seed, **kwargs)
        return jax.lax.stop_gradient(one_hot_sample - self.probs) + self.probs

    def logp(self, value: TensorType, **kwargs) -> TensorType:
        return self._dist.log_prob(value, **kwargs)

    def kl(self, other: distrax.DistributionLike, **kwargs) -> TensorType:
        if isinstance(other, RLlibToJaxDistribution):
            other = other._dist
        return self._dist.kl_divergence(other, **kwargs)

    @staticmethod
    def required_input_dim(space: gym.Space | gym.spaces.Discrete, **kwargs) -> int:  # noqa: ARG004
        return int(space.n)  # type: ignore[attr-defined]§

    def to_deterministic(self) -> Deterministic:
        if self.probs is not None:
            probs_or_logits = self.probs
        else:
            probs_or_logits = self.logits

        return Deterministic(loc=jnp.argmax(probs_or_logits, axis=-1))


if TYPE_CHECKING:
    Categorical(None, None)


class Deterministic(_EntropyArgCorrector, RLlibToJaxDistribution, distrax.Deterministic):
    def __init__(self, loc: chex.Array, atol: Optional[chex.Numeric] = None, rtol: Optional[chex.Numeric] = None):
        super().__init__(loc=loc, atol=atol, rtol=rtol)

    def _get_jax_distribution(
        self, loc: chex.Array, atol: Optional[chex.Numeric] = None, rtol: Optional[chex.Numeric] = None
    ) -> distrax.Distribution:
        self._dist: distrax.Deterministic
        return distrax.Deterministic(loc=loc, atol=atol, rtol=rtol)

    @staticmethod
    def required_input_dim(space: gym.Space, **kwargs) -> int:  # noqa: ARG004
        assert space.shape is not None
        return int(np.prod(space.shape, dtype=np.int32))

    def rsample(self, *, sample_shape=(), return_logp: bool = False, seed=None, **kwargs):
        raise NotImplementedError

    def to_deterministic(self) -> Self:
        return self


if TYPE_CHECKING:
    Deterministic(jnp.array(1))


# Note: Torch distribution works better with various structures
class MultiCategorical(RLlibToJaxDistribution):
    def __init__(
        self,
        categoricals: list[Categorical],
    ):
        RLlibToJaxDistribution.__init__(self)
        self._cats = categoricals

    def _get_jax_distribution(self) -> distrax.Distribution:
        self._dist: None
        return None

    def sample(self, *, seed, **kwargs) -> TensorType:
        if kwargs.get("return_logp", False):
            raise NotImplementedError("return_logp not implemented")
        arr: list[chex.Array] = [cat.sample(seed=seed) for cat in self._cats]  # pyright: ignore[reportAssignmentType]
        sample_ = jnp.stack(arr, axis=-1)
        return sample_

    def rsample(self, sample_shape=(), *seed, **kwargs):  # noqa: ARG002  # no torch/tf class uses sample shape here
        arr: list[chex.Array] = [cat.rsample(seed=seed) for cat in self._cats]  # type: ignore[attr-defined]
        sample_ = jnp.stack(arr, axis=-1)
        return sample_

    def logp(self, value: chex.Array, **kwargs) -> TensorType:  # noqa: ARG002
        actions = jnp.unstack(jnp.astype(value, jnp.int32), axis=-1)
        logps = jnp.stack([cat.logp(act) for cat, act in zip(self._cats, actions)])
        return jnp.sum(logps, axis=0)

    def entropy(self, **kwargs) -> TensorType:  # noqa: ARG002
        return jnp.sum(jnp.stack([cat.entropy() for cat in self._cats], axis=-1), axis=-1)

    def kl(self, other: MultiCategorical, **kwargs) -> TensorType:  # noqa: ARG002  # type: ignore[override]
        kls = jnp.stack([cat.kl(oth_cat) for cat, oth_cat in zip(self._cats, other._cats)], axis=-1)
        return jnp.sum(kls, axis=-1)

    @staticmethod
    def required_input_dim(space: gym.Space, **kwargs) -> int:  # noqa: ARG004
        # assert isinstance(space, gym.spaces.MultiDiscrete)
        return int(np.sum(space.nvec))  # type: ignore[attr-defined]

    @classmethod
    def from_logits(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls,
        logits: chex.Array,
        input_lens: list[int],
        **kwargs,  # noqa: ARG003
    ) -> "MultiCategorical":
        """Creates this Distribution from logits (and additional arguments).

        If you wish to create this distribution from logits only, please refer to
        `Distribution.get_partial_dist_cls()`.

        Args:
            logits: The tensor containing logits to be separated by logit_lens.
                child_distribution_cls_struct: A struct of Distribution classes that can
                be instantiated from the given logits.
            input_lens: A list of integers that indicate the length of the logits
                vectors to be passed into each child distribution.
            **kwargs: Forward compatibility kwargs.
        """
        categoricals = [Categorical(logits=logits) for logits in jnp.split(logits, input_lens, axis=-1)]

        return cls(categoricals=categoricals)

    def to_deterministic(self) -> "MultiDistribution":  # type: ignore
        return MultiDistribution([cat.to_deterministic() for cat in self._cats])

    def log_prob(self, value):
        return self.logp(value)

    def _sample_n(self, key: jax.Array, n: int) -> Any:
        raise NotImplementedError("use sample method instead")

    @property
    def event_shape(self) -> Tuple[int, ...]:
        return ()


if TYPE_CHECKING:
    MultiCategorical([Categorical(None, None), Categorical(None, None)])


class MultiDistribution(RLlibToJaxDistribution):
    """Action distribution that operates on multiple, possibly nested actions."""

    def __init__(
        self,
        child_distribution_struct: Union[tuple, list, dict],
    ):
        """Initializes a TfMultiDistribution object.

        Args:
            child_distribution_struct: Any struct
                that contains the child distribution classes to use to
                instantiate the child distributions from `logits`.
        """
        super().__init__()
        self._original_struct = child_distribution_struct
        self._flat_child_distributions: Sequence[RLlibToJaxDistribution] = tree.flatten(child_distribution_struct)

    def _get_jax_distribution(self, *args, **kwargs) -> None:  # noqa: ARG002
        self._dist: None
        return None  # noqa

    def rsample(
        self,
        *,
        sample_shape: Optional[tuple[int, ...]] = None,
        **kwargs,
    ) -> Union[TensorType, Tuple[TensorType, TensorType]]:
        rsamples = []
        for dist in self._flat_child_distributions:
            rsample = dist.rsample(sample_shape=sample_shape, **kwargs)  # pyright: ignore[reportArgumentType] super has implicit optional
            rsamples.append(rsample)

        rsamples = tree.unflatten_as(self._original_struct, rsamples)
        return rsamples

    def logp(self, value, **kwargs) -> TensorType:  # noqa: ARG002
        # Single tensor input (all merged).
        if isinstance(value, (jax.Array, np.ndarray)):
            split_indices = []
            for dist in self._flat_child_distributions:
                if isinstance(dist, Categorical):
                    split_indices.append(1)
                elif isinstance(dist, MultiCategorical):
                    split_indices.append(len(dist._cats))
                else:
                    sample: chex.Array = dist.sample(seed=4423)  # XXX
                    # Cover Box(shape=()) case.
                    if len(sample.shape) == 1:
                        split_indices.append(1)
                    else:
                        split_indices.append(jnp.shape(sample)[1])
            split_value = jnp.split(value, split_indices, axis=1)
        # Structured or flattened (by single action component) input.
        else:
            split_value = tree.flatten(value)

        def map_(val: chex.Array, dist: RLlibToJaxDistribution):
            # Remove extra dimension if present.
            if isinstance(dist, Categorical) and len(val.shape) > 1 and val.shape[-1] == 1:
                val = jnp.squeeze(val, axis=-1)

            return dist.logp(val)

        # Remove extra categorical dimension and take the logp of each
        # component.
        flat_logps = tree.map_structure(map_, split_value, self._flat_child_distributions)

        return sum(flat_logps)  # type: ignore

    def kl(self, other: MultiDistribution, **kwargs):  # pyright: ignore[reportIncompatibleMethodOverride]
        kl_list = [d.kl(o, **kwargs) for d, o in zip(self._flat_child_distributions, other._flat_child_distributions)]
        return sum(kl_list)

    def entropy(self, **kwargs):
        entropy_list = [d.entropy(**kwargs) for d in self._flat_child_distributions]
        return sum(entropy_list)

    def sample(self, *, seed, **kwargs):
        child_distributions_struct = tree.unflatten_as(self._original_struct, self._flat_child_distributions)
        return tree.map_structure(lambda s: s.sample(seed=seed, **kwargs), child_distributions_struct)

    @staticmethod
    def required_input_dim(space: gym.Space, input_lens: list[int], **kwargs) -> int:  # noqa: ARG004  # pyright: ignore[reportIncompatibleMethodOverride]
        return sum(input_lens)

    @classmethod
    def from_logits(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls,
        logits: chex.Array,
        child_distribution_cls_struct: Union[dict, Iterable],
        input_lens: Union[dict, list[int]],
        **kwargs,  # noqa: ARG003
    ) -> "MultiDistribution":
        """Creates this Distribution from logits (and additional arguments).

        If you wish to create this distribution from logits only, please refer to
        `Distribution.get_partial_dist_cls()`.

        Args:
            logits: The tensor containing logits to be separated by `input_lens`.
                child_distribution_cls_struct: A struct of Distribution classes that can
                be instantiated from the given logits.
            child_distribution_cls_struct: A struct of Distribution classes that can
                be instantiated from the given logits.
            input_lens: A list or dict of integers that indicate the length of each
                logit. If this is given as a dict, the structure should match the
                structure of child_distribution_cls_struct.
            space: The possibly nested output space.
            **kwargs: Forward compatibility kwargs.

        Returns:
            A TfMultiDistribution object.
        """
        logit_lens = tree.flatten(input_lens)
        child_distribution_cls_list: Sequence[RllibDistribution] = tree.flatten(child_distribution_cls_struct)
        split_logits = jnp.split(logits, logit_lens, axis=1)

        child_distribution_list = tree.map_structure(
            lambda dist, input_: dist.from_logits(input_),
            child_distribution_cls_list,
            list(split_logits),
        )

        child_distribution_struct: Any | dict | list = tree.unflatten_as(
            child_distribution_cls_struct, child_distribution_list
        )

        return MultiDistribution(
            child_distribution_struct=child_distribution_struct,
        )

    def to_deterministic(self) -> "MultiDistribution":  # pyright: ignore[reportIncompatibleMethodOverride]
        flat_deterministic_dists = [dist.to_deterministic for dist in self._flat_child_distributions]
        deterministic_dists: Any | dict | list = tree.unflatten_as(self._original_struct, flat_deterministic_dists)
        return MultiDistribution(deterministic_dists)

    def log_prob(self, value):
        return self.logp(value)

    @property
    def event_shape(self) -> Tuple[int, ...]:
        logger.warning("Event Shape Accessed")
        return ()

    def _sample_n(self, key: jax.Array, n: int) -> Any:
        raise NotImplementedError("use sample method instead")



from __future__ import annotations

import logging
from collections.abc import Callable
from typing import TYPE_CHECKING, Optional, cast

from ray.rllib.algorithms.callbacks import DefaultCallbacks
from ray.rllib.utils.metrics import EVALUATION_RESULTS

from ray_utilities.constants import EVALUATED_THIS_STEP
from ray_utilities.discrete_evaluation import discrete_evaluate_on_local_env_runner

if TYPE_CHECKING:
    from interpretable_ddts.rllib_port.ddt_ppo_module import DDTModule  # TODO: upstream
    from ray.rllib.algorithms import Algorithm
    from ray.rllib.env.single_agent_env_runner import SingleAgentEnvRunner
    from ray.rllib.policy.sample_batch import MultiAgentBatch, SampleBatch
    from ray.rllib.utils.metrics.metrics_logger import MetricsLogger
    from ray.rllib.utils.typing import ResultDict

__all__ = ["DiscreteEvalCallback"]

logger = logging.getLogger(__name__)

DiscreteEvalFunctionType = Callable[
    ["Algorithm", "SingleAgentEnvRunner", "MetricsLogger"],
    tuple["ResultDict | None", int, int, list["SampleBatch | MultiAgentBatch"]],
]


class DiscreteEvalCallback(DefaultCallbacks):
    _warned_once: bool = False

    def on_evaluate_end(
        self,
        *,
        algorithm: Algorithm,
        metrics_logger: Optional[MetricsLogger] = None,
        evaluation_metrics: dict,
        **kwargs,  # noqa: ARG002
    ) -> None:
        env_runner = algorithm.env_runner
        eval_workers = algorithm.eval_env_runner_group
        if eval_workers is None:  # type: ignore[comparison-overlap]
            env_runner = algorithm.env_runner_group.local_env_runner  # type: ignore[attr-defined]
        elif eval_workers.num_healthy_remote_workers() == 0:
            env_runner = algorithm.eval_env_runner
        elif eval_workers.local_env_runner:
            if not self._warned_once:
                logger.info(
                    "Parallel discrete evaluation not implemented. Using one local env_runner for discrete evaluation."
                )
                self._warned_once = True
            env_runner = eval_workers.local_env_runner
        else:
            if not self._warned_once:
                logger.warning("No eval workers available for discrete evaluation. Using local env_runner.")
                self._warned_once = True
            env_runner = algorithm.eval_env_runner
            # possibly still use eval_env_runner
            # raise NotImplementedError("Parallel discrete evaluation not implemented")
        env_runner = cast("SingleAgentEnvRunner", env_runner)
        if metrics_logger is None:
            logger.warning("No metrics logger provided for discrete evaluation")
            metrics_logger = algorithm.metrics
        module: DDTModule = env_runner.module  # type: ignore[assignment]
        evaluation_metrics[EVALUATED_THIS_STEP] = True  # Note: NotRequired key
        if not getattr(module, "CAN_USE_DISCRETE_EVAL", False):
            return
        assert metrics_logger
        module.switch_mode(discrete=True)
        assert module.is_discrete
        (
            discrete_eval_results,
            _env_steps,
            _agent_steps,
            _batches,
        ) = discrete_evaluate_on_local_env_runner(algorithm, env_runner, metrics_logger)
        module.switch_mode(discrete=False)
        assert module.is_discrete is False
        assert discrete_eval_results is None  # new API stack
        if discrete_eval_results is None:  # and algorithm.config.enable_env_runner_and_connector_v2:
            # FIXME: reduce changed in ray version, does not take arguments anympre
            try:
                discrete_eval_results = metrics_logger.reduce((EVALUATION_RESULTS, "discrete"), return_stats_obj=False)  # pyright: ignore[reportCallIssue]
            except TypeError:
                # new ray versions; untested, might reduce too much
                discrete_eval_results = metrics_logger.reduce()[EVALUATION_RESULTS]["discrete"]
        evaluation_metrics["discrete"] = discrete_eval_results



from __future__ import annotations

import gymnasium as gym
from ray_utilities.jax.distributions.jax_distributions import Categorical, Normal, RLlibToJaxDistribution

__all__ = ["GetJaxDistributionsMixin"]


class GetJaxDistributionsMixin:
    """
    Mixin providing methods for RLModules

    Methods:
        - get_train_action_dist_cls
        - get_exploration_action_dist_cls
        - get_inference_action_dict_cls
    """

    action_dist_cls: type[RLlibToJaxDistribution]
    action_space: gym.Space

    def get_train_action_dist_cls(self) -> type[RLlibToJaxDistribution]:
        return self.get_inference_action_dist_cls()

    def get_exploration_action_dist_cls(self) -> type[RLlibToJaxDistribution]:
        return self.get_inference_action_dist_cls()

    def get_inference_action_dist_cls(self) -> type[RLlibToJaxDistribution]:
        if self.action_dist_cls is not None:  # pyright: ignore[reportUnnecessaryComparison]
            return self.action_dist_cls
        if isinstance(self.action_space, gym.spaces.Discrete):
            return Categorical
        if isinstance(self.action_space, gym.spaces.Box):
            return Normal
        raise ValueError(
            f"Default action distribution for action space "
            f"{self.action_space} not supported! Either set the "
            f"`self.action_dist_cls` property in your RLModule's `setup()` method "
            f"to a subclass of `ray.rllib.models.torch.torch_distributions."
            f"TorchDistribution` or - if you need different distributions for "
            f"inference and training - override the three methods: "
            f"`get_inference_action_dist_cls`, `get_exploration_action_dist_cls`, "
            f"and `get_train_action_dist_cls` in your RLModule."
        )



from __future__ import annotations

import logging
import sys
from functools import partial
from typing import TYPE_CHECKING, Any, Optional, cast

from ray.rllib.callbacks.callbacks import RLlibCallback
from ray.rllib.utils.annotations import override
from ray.tune.result import SHOULD_CHECKPOINT
from typing_extensions import Self

from ray_utilities.training.helpers import get_current_step

if TYPE_CHECKING:
    from collections.abc import Callable

    from ray.rllib.algorithms.algorithm import Algorithm
    from ray.tune.experiment import Trial

    from ray_utilities.typing.algorithm_return import StrictAlgorithmReturnData
    from ray_utilities.typing.metrics import LogMetricsDict

_logger = logging.getLogger(__name__)


class AlgoMetricCheckpointer(RLlibCallback):
    """Callbacks that adds ``SHOULD_CHECKPOINT`` to results if a metric condition is met."""

    condition: Optional[Callable[[dict[str, Any]], bool]] = None

    @classmethod
    def make_callback_class(cls, **kwargs) -> type[Self]:
        return partial(cls, **kwargs)  # pyright: ignore[reportReturnType]

    def __init__(
        self, metric_name: Optional[str] = None, condition: Optional[Callable[[dict[str, Any]], bool]] = None
    ) -> None:
        if self.condition is None and condition is None:
            raise ValueError(
                "Condition must be provided for MetricCheckpointer. Either as class variable or in constructor."
            )
        if self.condition is not None and condition is not None:
            _logger.warning("Both class variable and constructor condition provided. Using constructor condition.")
        super().__init__()
        self.metric_name = metric_name or "Unknown"
        self.condition = self.condition or condition
        assert self.condition
        self._last_checkpoint_iteration = -1
        self._last_checkpoint_value = None

    def _set_checkpoint(
        self, result: StrictAlgorithmReturnData | LogMetricsDict, iteration: int, trial: Trial | None = None
    ) -> None:
        current_step = get_current_step(result)
        # config available in trial.config
        if self.condition(cast("dict[str, Any]", result)):  # pyright: ignore[reportOptionalCall]
            self._last_checkpoint_iteration = iteration
            self._last_checkpoint_value = result.get(self.metric_name, None)
            self._last_checkpoint_step = current_step
            result[SHOULD_CHECKPOINT] = True
            _logger.info(
                "Checkpointing trial %s at iteration %s, step %d with: metric %s = %s",
                trial.trial_id if trial else "",
                iteration,
                self._last_checkpoint_step,
                self.metric_name,
                self._last_checkpoint_value,
            )
        # else:
        #    _logger.debug(
        #        "Not checkpointing trial %s at iteration %s, step %d with: metric %s = %s",
        #        trial.trial_id if trial else "",
        #        iteration,
        #        self._last_checkpoint_step,
        #        self.metric_name,
        #        result.get(self.metric_name, None),
        #    )

    @override(RLlibCallback)
    def on_train_result(
        self,
        *,
        algorithm: "Algorithm",
        result: dict[str, Any],
        **kwargs,
    ) -> None:
        # NOTE: Algorithm iteration might be off by 1 as it is logged earlier in the training loop.
        self._set_checkpoint(result, iteration=algorithm.iteration)  # pyright: ignore[reportArgumentType]


class AlgoStepCheckpointer(AlgoMetricCheckpointer):  # type: ignore
    """Checkpoints trials based on a specific metric condition."""

    def _condition(self, result: StrictAlgorithmReturnData | LogMetricsDict | dict) -> bool:
        steps_since_last_checkpoint = get_current_step(result) - self._last_checkpoint_step  # pyright: ignore[reportArgumentType]

        return steps_since_last_checkpoint >= self._checkpoint_frequency

    def __init__(self, checkpoint_frequency: int = 50_000) -> None:
        if checkpoint_frequency == 0:
            _logger.info("Checkpoint frequency is set to 0, disabling step checkpointing.")
            checkpoint_frequency = sys.maxsize
        self._checkpoint_frequency = checkpoint_frequency
        super().__init__("current_step", self._condition)

    @classmethod
    def make_callback_class(cls, *, checkpoint_frequency, **kwargs) -> type[Self]:
        return partial(cls, checkpoint_frequency=checkpoint_frequency, **kwargs)  # pyright: ignore[reportReturnType]



"""Setup classes for Ray RLlib experiments and hyperparameter tuning.

The setup framework is the core of Ray Utilities, providing a comprehensive system
for configuring, running, and managing reinforcement learning experiments. It integrates
Ray RLlib with Ray Tune for scalable hyperparameter optimization while maintaining
clean separation between experiment configuration and execution.

**Core Design Principles:**

- **Configuration Management**: Centralized handling of algorithm configs, environment
  settings, and hyperparameters through type-safe argument parsing
- **Experiment Lifecycle**: Complete management from initialization through training
  to result collection and experiment tracking
- **Hyperparameter Optimization**: Native integration with Ray Tune for distributed
  parameter search with intelligent defaults

**Main Components:**

- :class:`ExperimentSetupBase`: Abstract foundation for all experiment configurations.
  Handles argument parsing, config validation, and trainable creation
- :class:`PPOSetup`: Production-ready PPO algorithm setup with sensible defaults
  and extensive customization options
- :class:`TunerSetup`: Ray Tune integration for hyperparameter optimization with
  support for various schedulers, search algorithms, and stopping criteria

**Typical Workflow:**

1. **Setup Creation**: Instantiate a setup class (e.g., :class:`PPOSetup`)
2. **Configuration**: Modify algorithm config, environment, and hyperparameters
3. **Execution**: Run experiments via :func:`~ray_utilities.runfiles.run_tune.run_tune`
4. **Analysis**: Process results using Ray Tune's analysis tools

**Examples:**

Basic experiment setup:

>>> from ray_utilities.setup import PPOSetup
>>> setup = PPOSetup()
>>> setup.config.env = "CartPole-v1"
>>> setup.config.lr = 0.001

Hyperparameter optimization:

>>> setup.add_tune_config({"lr": tune.grid_search([0.001, 0.01, 0.1]), "gamma": tune.uniform(0.9, 0.99)})

Custom environment with training configuration:

>>> setup.config.env = "MyCustomEnv"
>>> setup.config.training_iteration = 100
>>> setup.config.checkpoint_freq = 10

See Also:
    :mod:`ray_utilities.runfiles.run_tune`: Main execution functions
    :class:`ray.rllib.algorithms.ppo.PPOConfig`: Underlying algorithm configuration
    :class:`ray.tune.Tuner`: Ray Tune hyperparameter optimization
"""

from .algorithm_setup import PPOSetup
from .experiment_base import ExperimentSetupBase
from .tuner_setup import TunerSetup

__all__ = ["ExperimentSetupBase", "PPOSetup", "TunerSetup"]



from ast import literal_eval
from typing import Optional

from ray.rllib.core.rl_module.default_model_config import DefaultModelConfig
from tap import Tap
import tree

from ray_utilities.config import DefaultArgumentParser

try:
    from frozenlist import FrozenList  # pyright: ignore[reportAssignmentType]
except ImportError:
    from copy import deepcopy

    class FrozenList(list):
        def freeze(self):
            pass

        def append(self, *args, **kwargs) -> None:  # noqa: ARG002
            raise RuntimeError("Cannot modify frozen list.")

        insert = pop = append
        frozen: bool = True

        def __deepcopy__(self, memo):
            # Return a new FrozenList with the same elements
            copied_list = FrozenList((deepcopy(item, memo) for item in self))
            if self.frozen:
                copied_list.freeze()
            return copied_list

    _fcnet_hiddens_default = FrozenList([256, 256])
    _head_fcnet_hiddens_default = FrozenList([])
else:
    _fcnet_hiddens_default = FrozenList([256, 256])
    _fcnet_hiddens_default.freeze()
    _head_fcnet_hiddens_default = FrozenList([])
    _head_fcnet_hiddens_default.freeze()

__all__ = ["MLPArgumentParser"]


class SimpleMLPParser(Tap):
    """Keys must align with RLlib's DefaultModelConfig"""

    # Documentation taken from ray's DefaultModelConfig

    fcnet_hiddens: list[int] = _fcnet_hiddens_default
    """
    List containing the sizes (number of nodes) of a fully connected (MLP) stack.
    Note that in an encoder-based default architecture with a policy head (and
    possible value head), this setting only affects the encoder component. To set the
    policy (and value) head sizes, use `post_fcnet_hiddens`, instead. For example,
    if you set `fcnet_hiddens=[32, 32]` and `post_fcnet_hiddens=[64]`, you would get
    an RLModule with a [32, 32] encoder, a [64, act-dim] policy head, and a [64, 1]
    value head (if applicable).

    Passing a list as a string: --fcnet_hiddens="[256, 256]" is equivalent to --fcnet_hiddens 256 256
    """

    fcnet_activation: str = DefaultModelConfig.fcnet_activation  # tanh
    """
    Activation function descriptor for the stack configured by `fcnet_hiddens`.
    Supported values are: 'tanh', 'relu', 'swish' (or 'silu', which is the same),
    and 'linear' (or None).
    """

    fcnet_kernel_initializer: Optional[str] = None
    """
    Initializer function descriptor for the weight/kernel matrices in the stack
    configured by `fcnet_hiddens`. Supported values are the names of initializers
    as strings. See https://pytorch.org/docs/stable/nn.init.html for `torch`. If
    `None` (default), the default initializer defined by `torch` is used.
    """

    fcnet_bias_initializer: Optional[str] = None
    """
    Initializer function descriptor for the bias vectors in the stack
    configured by `fcnet_hiddens`. Supported values are the names of initializers
    as strings. See https://pytorch.org/docs/stable/nn.init.html for `torch`. If
    `None` (default), the default initializer defined by `torch` is used.
    """

    # ====================================================
    # Head configs (e.g. policy- or value function heads)
    # ====================================================
    head_fcnet_hiddens: list[int] = _head_fcnet_hiddens_default
    """
    List containing the sizes (number of nodes) of a fully connected (MLP) head
    (e.g., policy-, value-, or Q-head). To configure the encoder architecture,
    use `fcnet_hiddens` instead.

    Passing a list string: --head_fcnet_hiddens="[256, 256]" is equivalent to --head_fcnet_hiddens 256 256

    Default:
        Empty list; no layers
    """

    head_fcnet_activation: str = DefaultModelConfig.head_fcnet_activation  # relu
    """
    Activation function descriptor for the stack configured by `head_fcnet_hiddens`.
    Supported values are: 'tanh', 'relu', 'swish' (or 'silu', which is the same),
    and 'linear' (or None).
    """

    head_fcnet_kernel_initializer: Optional[str] = None
    """
    Initializer function descriptor for the weight/kernel matrices in the stack
    configured by `head_fcnet_hiddens`. Supported values are the names of initializers
    as strings. See https://pytorch.org/docs/stable/nn.init.html for `torch`. If `None`
    (default), the default initializer defined by `torch` is used.
    """

    head_fcnet_bias_initializer: Optional[str] = None
    """
    Initializer function descriptor for the bias vectors in the stack configured
    by `head_fcnet_hiddens`. Supported values are the names of initializers as strings.
    See https://pytorch.org/docs/stable/nn.init.html for `torch`. If `None` (default),
    the default initializer defined by `torch` is used.
    """

    vf_share_layers: bool = DefaultModelConfig.vf_share_layers  # True
    """
    Whether encoder layers (defined by `fcnet_hiddens` or `conv_filters`) should be
    shared between policy- and value function.
    """

    def configure(self) -> None:
        super().configure()
        self.add_argument("--fcnet_hiddens", nargs="*", type=literal_eval)
        self.add_argument("--head_fcnet_hiddens", nargs="*", type=literal_eval)

    def process_args(self) -> None:
        # flatten list of lists
        self.fcnet_hiddens = tree.flatten(self.fcnet_hiddens)
        self.head_fcnet_hiddens = tree.flatten(self.head_fcnet_hiddens)
        super().process_args()


class MLPArgumentParser(SimpleMLPParser, DefaultArgumentParser):
    pass



from __future__ import annotations

from typing import TYPE_CHECKING, Final, Protocol

from typing_extensions import NotRequired, Required, TypedDict

from .algorithm_return import _NotRequiredEnvRunners

if TYPE_CHECKING:
    from .metrics import _LogMetricsEvaluationResultsDict

# pyright: enableExperimentalFeatures=true


class TrainableReturnData(_NotRequiredEnvRunners, total=False, closed=False):
    """The return type of the trainable's step method."""

    evaluation: Required[_LogMetricsEvaluationResultsDict]
    training_iteration: int
    done: Required[bool]
    current_step: Required[int]
    comment: str
    trial_id: int | str


# Further


class RewardsDict(TypedDict):
    """A dictionary containing the rewards for the current training step."""

    running_reward: float
    running_eval_reward: float
    eval_mean: float
    disc_eval_reward: NotRequired[float]
    disc_eval_mean: NotRequired[float]
    disc_running_eval_reward: NotRequired[float]


class RewardUpdater(Protocol):
    """A protocol for a partial function that updates rewards."""

    def __call__(self, new_reward: float) -> float: ...

    keywords: Final[TypedDict[{"reward_array": list[float]}]]


class RewardUpdaters(TypedDict, extra_items=RewardUpdater):
    """A dictionary containing the reward updaters for the current training step."""

    # The keys are the names of the reward updaters, and the values are the functions that update the rewards.
    # The functions take a float as input and return a float.

    running_reward: RewardUpdater
    eval_reward: RewardUpdater
    disc_eval_reward: NotRequired[RewardUpdater]



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from ray.rllib.utils.metrics import ENV_RUNNER_RESULTS

from ray_utilities.constants import EPISODE_METRICS_KEYS

if TYPE_CHECKING:
    from ray.rllib.algorithms.algorithm import Algorithm
    from ray.rllib.env.env_runner import EnvRunner
    from ray.rllib.utils.metrics.metrics_logger import MetricsLogger


def _reset_episode_metrics(runner: EnvRunner):
    for key in EPISODE_METRICS_KEYS:
        runner.metrics.set_value(key, [])


def reset_eval_episode_metrics_each_time(
    algorithm: "Algorithm",
    **kwargs,
):
    """
    RLlib used metrics_num_episodes_for_smoothing to metrics like episode_reward_mean over
    a window of episodes.
    But when there are more episodes than the smoothing window, the result of the first episodes is
    ignored.
    Likewise if there are less episodes, information from the last iteration bleeds in.

    This callback is used on the algorithm `on_evaluation_start` (only eval)
    """
    algorithm.eval_env_runner_group.foreach_env_runner(lambda env_runner: _reset_episode_metrics(env_runner))


def reset_episode_metrics_each_iteration(
    env_runner: Optional["EnvRunner"] = None,
    metrics_logger: Optional[MetricsLogger] = None,
    **kwargs,
) -> None:
    """
    RLlib used metrics_num_episodes_for_smoothing to metrics like episode_reward_mean over
    a window of episodes.
    But when there are more episodes than the smoothing window, the result of the first episodes is
    ignored.
    Likewise if there are less episodes, information from the last iteration bleeds in.

    This callback is used on the algorithm `on_sample_end` (all env runners)
    """
    if metrics_logger is None:
        assert env_runner
        metrics_logger = env_runner.metrics
    for key in EPISODE_METRICS_KEYS:  # keys are not present in the first iteration
        metrics_logger.delete(key, key_error=metrics_logger.peek("weight_seq_no", compile=True, default=0))  # pyright: ignore[reportArgumentType]



"""
Note:
   Creation of loggers is done in _create_default_callbacks which check for inheritance from Callback class.

"""

from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any

from ray import cloudpickle
from ray.air.constants import EXPR_RESULT_FILE
from ray.tune.experiment.trial import Trial
from ray.tune.logger import JsonLoggerCallback
from ray.tune.utils.util import SafeFallbackEncoder  # pyright: ignore[reportPrivateImportUsage]

from ray_utilities.callbacks.tuner._file_logger_fork_mixin import FileLoggerForkMixin
from ray_utilities.callbacks.tuner.new_style_logger_callback import NewStyleLoggerCallback
from ray_utilities.constants import FORK_FROM
from ray_utilities.postprocessing import remove_videos

if TYPE_CHECKING:
    from ray.tune.experiment.trial import Trial

    from ray_utilities.typing.metrics import AnyLogMetricsDict


logger = logging.getLogger(__name__)


class AdvJsonLoggerCallback(NewStyleLoggerCallback, FileLoggerForkMixin, JsonLoggerCallback):
    """Logs trial results in JSON format.

    Also writes to a results file and param.json file when results or
    configurations are updated. Experiments must be executed with the
    JsonLoggerCallback to be compatible with the ExperimentAnalysis tool.

    This updated class does not log videos stored in the :const:`DEFAULT_VIDEO_DICT_KEYS`.

    When a trial is forked (has ``FORK_FROM`` in config), creates a new JSON file
    and optionally copies data from the parent trial.
    """

    def _get_file_extension(self) -> str:
        return "json"

    def _get_file_base_name(self) -> str:
        return "result"

    def _get_default_file_name(self) -> str:
        return EXPR_RESULT_FILE

    def update_config(self, trial: "Trial", config: dict):
        # Overwrite super method to not use default names.
        # We can call super if this trial is not forked and has never been forked before.
        if not self.is_trial_forked(trial) and FORK_FROM not in trial.config:
            # Use default names
            super().update_config(trial, config)
            return
        self._trial_configs[trial] = config

        trial_local_path = Path(trial.local_path)  # pyright: ignore[reportArgumentType]

        param_file_name_base = f"params-fork-{self.make_forked_trial_id(trial, trial.config[FORK_FROM])}"

        config_out = Path(trial_local_path, param_file_name_base + ".json")
        with config_out.open("w") as f:
            json.dump(
                self._trial_configs[trial],
                f,
                indent=2,
                sort_keys=True,
                cls=SafeFallbackEncoder,
            )

        config_pkl = Path(trial_local_path, param_file_name_base + ".pkl")
        with config_pkl.open("wb") as f:
            cloudpickle.dump(self._trial_configs[trial], f)  # pyright: ignore[reportPrivateImportUsage]

    def _setup_file_handle(self, trial: Trial, local_file_path: Path) -> None:
        """Open the JSON file handle and update config."""
        trial.init_local_path()
        self.update_config(trial, trial.config)
        self._trial_files[trial] = local_file_path.open("at")

    def _handle_missing_parent_file(self, trial: Trial, local_file_path: Path) -> None:
        """Handle missing parent file for JSON logger."""
        logger.warning(
            "Trial %s forked but found no logfile for parent, starting fresh .json log file: %s",
            trial.trial_id,
            local_file_path,
        )

    def log_trial_result(self, iteration: int, trial: Trial, result: dict[str, Any] | AnyLogMetricsDict):
        super().log_trial_result(
            iteration,
            trial,
            remove_videos(result),
        )

    def log_trial_start(self, trial: Trial):
        # NOTE: Because JsonLoggerCallback.log_trial_start is not compatible with forked trials
        # set _call_super_log_trial_start to False to not call it in this chain.
        call_super = True
        if FORK_FROM in trial.config or self.is_trial_forked(trial):
            self._call_super_log_trial_start = call_super = False
        super().log_trial_start(trial)
        assert self._call_super_log_trial_start is True
        if not call_super:
            # As we skip the class after TrackForkedTrialsMixin, continue after the original base.
            super(JsonLoggerCallback, self).log_trial_start(trial)


if TYPE_CHECKING:  # Check ABC
    AdvJsonLoggerCallback()



"""Algorithm configuration creation utilities for Ray RLlib experiments.

This module provides comprehensive utilities for creating and configuring Ray RLlib
algorithm configurations based on parsed command-line arguments and experiment
specifications. It handles the complex setup of algorithms with proper environment
configuration, callback registration, learner setup, and module specifications.

The main function :func:`create_algorithm_config` serves as a centralized point
for translating experiment parameters into properly configured RLlib algorithm
instances, supporting both the new and legacy RLlib APIs.

Key Components:
    - :func:`create_algorithm_config`: Main configuration creation function
    - Environment setup with seeding and rendering support
    - Callback system integration and management
    - Learner class configuration and mixing
    - RL module and catalog class handling

The module bridges the gap between high-level experiment specifications and
low-level RLlib configuration requirements, providing a clean interface for
complex algorithm setup scenarios.
"""

from __future__ import annotations

import logging
import sys
from typing import TYPE_CHECKING, Any, Final, Literal, Optional, TypeVar, cast

from ray_utilities.callbacks.algorithm.dynamic_evaluation_callback import DynamicEvalInterval
from ray_utilities.callbacks.algorithm.model_config_saver_callback import save_model_config_and_architecture
from ray_utilities.warn import (
    warn_about_larger_minibatch_size,
    warn_if_batch_size_not_divisible,
    warn_if_minibatch_size_not_divisible,
)

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup
import gymnasium as gym
from gymnasium.envs.registration import VectorizeMode
from ray.rllib.algorithms.callbacks import DefaultCallbacks, make_multi_callbacks
from ray.rllib.algorithms.ppo.ppo import PPOConfig
from ray.rllib.core.rl_module import RLModule, RLModuleSpec
from ray.tune import logger as tune_logger

from ray_utilities.callbacks.algorithm.discrete_eval_callback import DiscreteEvalCallback
from ray_utilities.callbacks.algorithm.env_render_callback import make_render_callback
from ray_utilities.callbacks.algorithm.exact_sampling_callback import exact_sampling_callback
from ray_utilities.callbacks.algorithm.seeded_env_callback import SeedEnvsCallback
from ray_utilities.config import add_callbacks_to_config, seed_environments_for_config
from ray_utilities.learners import mix_learners
from ray_utilities.learners.leaner_with_debug_connector import LearnerWithDebugConnectors

if TYPE_CHECKING:
    from ray.rllib.algorithms import AlgorithmConfig
    from ray.rllib.core.learner import Learner
    from ray.rllib.core.models.catalog import Catalog

    from ray_utilities.config import DefaultArgumentParser
    from ray_utilities.setup.experiment_base import NamespaceType
    from ray_utilities.typing.generic_rl_module import CatalogWithConfig, RLModuleWithConfig

_ConfigType = TypeVar("_ConfigType", bound="AlgorithmConfig")

_ModelConfig = TypeVar("_ModelConfig", bound="None | dict | Any")

logger = logging.getLogger(__name__)


def create_algorithm_config(
    args: dict[str, Any] | NamespaceType[DefaultArgumentParser],
    env_type: Optional[str | gym.Env] = None,
    env_seed: Optional[int] = None,
    *,
    new_api: Optional[bool] = True,
    module_class: Optional[type[RLModule | RLModuleWithConfig[_ModelConfig]]],
    catalog_class: Optional[type[Catalog | CatalogWithConfig[_ModelConfig]]],
    learner_class: Optional[type["Learner"]] = None,
    model_config: dict[str, Any] | _ModelConfig,
    config_class: Optional[type[_ConfigType]] = PPOConfig,
    framework: Literal["torch", "tf2"] | Any,
    base_config: Optional[_ConfigType] = None,
    discrete_eval: bool = False,
) -> tuple[_ConfigType, RLModuleSpec]:
    """Create a comprehensive Ray RLlib algorithm configuration from experiment parameters.

    This function provides a centralized way to create properly configured RLlib
    algorithm instances based on parsed command-line arguments and experiment
    specifications. It handles environment setup, callback registration, learner
    configuration, and RL module specifications.

    The function supports both new and legacy RLlib APIs and provides extensive
    customization options for advanced use cases while maintaining sensible
    defaults for common scenarios.

    Args:
        args: Parsed command-line arguments containing experiment configuration.
            Can be a dictionary or a :class:`DefaultArgumentParser` instance.
        env_type: Environment identifier or Gymnasium environment instance.
            If not provided, will be extracted from args.
        env_seed: Environment seed for reproducibility. Deprecated in favor of
            :class:`~ray_utilities.callbacks.algorithm.seeded_env_callback.SeedEnvsCallback`.
        new_api: Whether to use the new RLlib API (RLModule/Learner). Defaults to ``True``.
        module_class: Custom RL module class for the algorithm. Should inherit from
            :class:`ray.rllib.core.rl_module.RLModule`. If ``None``, requires manual
            configuration afterward.
        catalog_class: Catalog class used with the ``module_class`` for model creation.
            Should inherit from :class:`ray.rllib.core.models.catalog.Catalog`.
        learner_class: Custom learner class for training. If ``None``, uses algorithm defaults.
        model_config: Configuration dictionary for the neural network model implementation.
            Structure depends on the chosen framework and module class.
        config_class: RLlib algorithm configuration class. Defaults to
            :class:`ray.rllib.algorithms.ppo.PPOConfig`.
        framework: Deep learning framework to use (``"torch"`` or ``"tf2"``).
        base_config: Optional existing configuration instance to update instead of creating new.
        discrete_eval: Whether to add discrete evaluation capabilities through
            :class:`~ray_utilities.callbacks.algorithm.discrete_eval_callback.DiscreteEvalCallback`.

    Returns:
        A tuple containing:
            - **config**: Configured RLlib algorithm configuration instance
            - **module_spec**: RL module specification for the algorithm

    Raises:
        ExceptionGroup: When neither ``base`` nor ``config_class`` is provided,
            or when ``config_class`` is not a proper type.

    Example:
        >>> args = parser.parse_args(["--env", "CartPole-v1", "--framework", "torch"])
        >>> config, module_spec = create_algorithm_config(
        ...     args,
        ...     module_class=MyRLModule,
        ...     catalog_class=MyCatalog,
        ...     model_config={"hidden_layers": [256, 256]},
        ...     framework="torch",
        ... )

    Note:
        The function automatically configures environment seeding, callback registration,
        and evaluation settings based on the provided arguments. For discrete evaluation,
        ensure the environment supports the required interface.

    See Also:
        :class:`DefaultArgumentParser`: Argument parsing
        :class:`~ray_utilities.callbacks.algorithm.seeded_env_callback.SeedEnvsCallback`: Environment seeding
        :class:`~ray_utilities.callbacks.algorithm.discrete_eval_callback.DiscreteEvalCallback`: Discrete evaluation
    """
    if not base_config and not isinstance(config_class, type):
        raise ExceptionGroup(
            "base or config_class must be provided",
            [
                ValueError("base or config_class must be provided"),
                TypeError(f"config_class must be a type, not {type(config_class)}"),
            ],
        )
    if not isinstance(args, dict):
        if hasattr(args, "as_dict"):  # Tap
            args = cast("dict[str, Any]", args.as_dict())
        else:
            args = vars(args).copy()
    if not env_type and not args["env_type"]:
        raise ValueError("No environment specified")
    env_spec: Final = env_type or args["env_type"]
    del env_type
    assert env_spec, "No environment specified"
    if base_config:
        config = base_config
        if config_class and not issubclass(config_class, type(base_config)):
            logger.warning(
                "base config of type %s is not a subclass of config_class %s, "
                "change config_class or set config_class to None to avoid this warning.",
                type(base_config),
                config_class,
            )
    else:
        assert config_class is not None
        config = config_class()

    env_config: dict[str, Any] = {}  # kwargs for environment __init__
    if args["render_mode"]:
        env_config["render_mode"] = args["render_mode"]
    if env_seed is not None:
        logger.warning(
            "env_seed is deprecated, use SeedEnvsCallback/seed_environments_for_config(config, env_seed) instead, "
            "env creation might fail."
        )
        env_config.update({"seed": env_seed, "env_type": env_spec})
        # Will use a SeededEnvCallback to apply seed and generators
    config.environment(env_spec, env_config=env_config)
    if args["test"]:
        # increase time in case of debugging the sampler
        config.env_runners(sample_timeout_s=1000)
    try:
        config.env_runners(
            # experimental
            gym_env_vectorize_mode=(VectorizeMode.ASYNC if args["num_envs_per_env_runner"] > 1 else VectorizeMode.SYNC),  # pyright: ignore[reportArgumentType]
        )
    except TypeError:
        logger.error("Current ray version does not support AlgorithmConfig.env_runners(gym_env_vectorize_mode=...)")
    config.resources(
        # num_gpus=1 if args["gpu"] else 0,4
        # process that runs Algorithm.training_step() during Tune
        num_cpus_for_main_process=1,
        # num_learner_workers=4 if args["parallel"] else 1,
        # num_cpus_per_learner_worker=1,
        # num_cpus_per_worker=1,
    )
    config.env_runners(
        num_env_runners=2 if args["parallel"] and args["num_env_runners"] < 2 else args["num_env_runners"],
        num_cpus_per_env_runner=1,  # num_cpus_per_worker
        # How long an rollout episode lasts, for "auto" calculated from batch_size
        # total_train_batch_size / (num_envs_per_env_runner * num_env_runners)
        # rollout_fragment_length=1,  # Default: "auto"
        num_envs_per_env_runner=args["num_envs_per_env_runner"],
        # validate_env_runners_after_construction=args["test"],
        # 1) "truncate_episodes": Each call to `EnvRunner.sample()` returns a
        #    batch of at most `rollout_fragment_length * num_envs_per_env_runner` in
        #    size. The batch is exactly `rollout_fragment_length * num_envs`
        #    in size if postprocessing does not change batch sizes.
        # Use if not using GAE
        # 2) "complete_episodes": Each call to `EnvRunner.sample()` returns a
        #    batch of at least `rollout_fragment_length * num_envs_per_env_runner` in
        #    size. Episodes aren't truncated, but multiple episodes
        #    may be packed within one batch to meet the (minimum) batch size.
        batch_mode="truncate_episodes",
    )
    config.learners(
        # for fractional GPUs, you should always set num_learners to 0 or 1
        num_learners=1 if args["parallel"] else 0,
        num_cpus_per_learner=0 if args["test"] and args["num_jobs"] < 2 else 1,
        num_gpus_per_learner=1 if args["gpu"] else 0,
    )
    config.framework(framework)
    learner_mix: list[type[Learner]] = [learner_class or config.learner_class]
    if not args.get("keep_masked_samples", False):
        from ray_utilities.learners.remove_masked_samples_learner import RemoveMaskedSamplesLearner  # noqa: PLC0415

        learner_mix.insert(0, RemoveMaskedSamplesLearner)
    if False:  # NOTE: Must always be the first in the mix
        learner_mix.insert(0, LearnerWithDebugConnectors)
    learner_config_dict = {
        "dynamic_buffer": args["dynamic_buffer"],
        "dynamic_batch": args["dynamic_batch"],
        "total_steps": args["total_steps"],
        "remove_masked_samples": not args["keep_masked_samples"],
        "min_dynamic_buffer_size": args["min_step_size"],
        "max_dynamic_buffer_size": args["max_step_size"],
        "accumulate_gradients_every": args["accumulate_gradients_every"],
    }
    if len(learner_mix) > 1:
        mixed_learner_class = mix_learners(learner_mix)
        try:  # new upcoming interface  # TODO: Check when this is changed ray 2.50+
            config.learners(learner_class=mixed_learner_class)  # pyright: ignore[reportCallIssue]
        except TypeError:  # old interface
            config.training(learner_class=mixed_learner_class)
    try:  # new upcoming interface
        config.learners(learner_config_dict=learner_config_dict)  # pyright: ignore[reportCallIssue]
    except TypeError:  # old interface
        config.training(learner_config_dict=learner_config_dict)

    config.training(
        gamma=0.99,
        # with a growing number of Learners and to increase the learning rate as follows:
        # lr = [original_lr] * ([num_learners] ** 0.5)
        lr=args["lr"],
        # The total effective batch size is then
        # `num_learners` x `train_batch_size_per_learner` and you can
        # access it with the property `AlgorithmConfig.total_train_batch_size`.
        train_batch_size_per_learner=args["train_batch_size_per_learner"],
        grad_clip=0.5,
    )
    try:
        cast("PPOConfig", config).training(
            minibatch_size=args["minibatch_size"],
        )
    except TypeError:
        cast("PPOConfig", config).training(
            sgd_minibatch_size=args["minibatch_size"],
        )
    if isinstance(config, PPOConfig):
        config.training(
            # PPO Specific
            use_critic=True,
            clip_param=0.2,
            # grad_clip_by="norm",
            entropy_coeff=0.01,
            # vf_clip_param=10,
            use_kl_loss=False,
            use_gae=True,  # Must be true to use "truncate_episodes"
        )
    # Create a single agent RL module spec.
    # NOTE: This might needs adjustment when using VectorEnv
    if isinstance(config.env, str) and config.env != "seeded_env":
        init_env = gym.make(config.env)
    elif config.env == "seeded_env":
        if isinstance(env_spec, str):
            init_env = gym.make(env_spec)
        else:
            init_env = env_spec
    else:
        assert not TYPE_CHECKING or config.env
        init_env = gym.make(config.env.unwrapped.spec.id)  # pyright: ignore[reportOptionalMemberAccess]
    # Note: legacy keys are updated below
    module_spec = RLModuleSpec(
        module_class=module_class,
        observation_space=init_env.observation_space,
        action_space=init_env.action_space,
        model_config=cast("dict[str, Any]", model_config),
        catalog_class=catalog_class,
    )
    # module = module_spec.build()
    config.rl_module(  # only in RLModuleSpec is not sufficient
        rl_module_spec=module_spec,
    )
    if model_config is not None:
        config.rl_module(model_config=model_config)
    # https://docs.ray.io/en/latest/rllib/package_ref/doc/ray.rllib.algorithms.algorithm_config.AlgorithmConfig.evaluation.html
    evaluation_duration = 30
    config.evaluation(
        evaluation_interval=16,  # Note can be adjusted dynamically by DynamicEvalInterval
        evaluation_duration=evaluation_duration,
        evaluation_duration_unit="episodes",
        evaluation_num_env_runners=(
            2 if args["parallel"] and args["evaluation_num_env_runners"] < 2 else args["evaluation_num_env_runners"]
        ),
        # NOTE: Policy gradient algorithms are able to find the optimal
        # policy, even if this is a stochastic one. Setting "explore=False" here
        # results in the evaluation workers not using this optimal policy!
        evaluation_config=PPOConfig.overrides(
            explore=False,
            num_envs_per_env_runner=min(5, args["num_envs_per_env_runner"]),
            metrics_num_episodes_for_smoothing=evaluation_duration,  # take metrics over all eval episodes
        ),
    )
    # Stateless callbacks
    if not args["no_exact_sampling"]:
        add_callbacks_to_config(config, on_sample_end=exact_sampling_callback)
    add_callbacks_to_config(config, on_algorithm_init=save_model_config_and_architecture)
    # add_callbacks_to_config(config, on_sample_end=reset_episode_metrics_each_iteration)

    # region Stateful callbacks
    callbacks: list[type[DefaultCallbacks]] = []
    base_callbacks = None
    if base_config is not None and base_config.callbacks_class:
        if isinstance(base_config.callbacks_class, list):
            base_callbacks = base_config.callbacks_class
        else:
            base_callbacks = [base_config.callbacks_class]
    if discrete_eval:
        callbacks.append(DiscreteEvalCallback)
    if args["env_seeding_strategy"] == "sequential":
        # Must set this in the trainable with seed_environments_for_config(config, env_seed)
        logger.info(
            "Using sequential env seed strategy, "
            "Remember to call seed_environments_for_config(config, env_seed) with a seed acquired from the trial."
        )
    elif args["env_seeding_strategy"] == "same":
        # TODO: could use env_seed here, allows to sample a constant random seed != args["seed"]
        seed_environments_for_config(config, args["seed"])
    elif args["env_seeding_strategy"] == "constant":
        seed_environments_for_config(config, SeedEnvsCallback.env_seed)
    if args["render_mode"]:
        callbacks.append(make_render_callback())
    if not args["no_dynamic_eval_interval"]:
        callbacks.append(DynamicEvalInterval)
    if base_callbacks:
        same = set(base_callbacks).intersection(set(callbacks))
        only_in_base = set(base_callbacks).difference(set(callbacks))
        only_in_new = set(callbacks).difference(set(base_callbacks))
        if only_in_base or only_in_new:
            logger.warning(
                "Callbacks of base differs from callbacks in new config:\n%s vs\n%s", only_in_base, only_in_new
            )
        callbacks = [*(cb for cb in base_callbacks if cb not in same), *callbacks]
    if callbacks:
        if len(callbacks) == 1:
            callback_class = callbacks[0]
        else:
            callback_class = callbacks
        if False:
            # OLD API
            multi_callback = make_multi_callbacks(callback_class)
            # Necessary patch for new_api, cannot use this callback with new API
            multi_callback.on_episode_created = DefaultCallbacks.on_episode_created
            config.callbacks(callbacks_class=multi_callback)
        else:
            config.callbacks(callbacks_class=callback_class)
    # endregion

    config.reporting(
        keep_per_episode_custom_metrics=True,  # If True calculate max min mean
        log_gradients=args["test"],  # Default is True
        # Will smooth metrics in the reports, e.g. tensorboard
        # NOTE This value will smooth over num_episodes, which might be from the past iterations.
        # But should be > 1 to smooth over episodes from the current iteration
        # We use the reset_episode_metrics_each_iteration, so that no bleed over happens,
        # but for all episodes to be considered we use a high value here
        # Should use ~50-60 for a higher batch_size
        metrics_num_episodes_for_smoothing=30,  # Default is 100
    )
    config.debugging(
        # https://docs.ray.io/en/latest/rllib/package_ref/doc/ray.rllib.algorithms.algorithm_config.AlgorithmConfig.debugging.html#ray-rllib-algorithms-algorithm-config-algorithmconfig-debugging
        seed=args["seed"],
        log_sys_usage=False,
        # These loggers will log more metrics which are stored less-accessible in the ~/ray_results/logdir
        # Using these could be useful if no Tuner is used
        logger_config={"type": tune_logger.NoopLogger},
    )
    if new_api is not None:
        config.api_stack(enable_rl_module_and_learner=new_api, enable_env_runner_and_connector_v2=new_api)
    # Checks
    warn_about_larger_minibatch_size(
        minibatch_size=config.minibatch_size, train_batch_size_per_learner=config.train_batch_size_per_learner
    )
    warn_if_batch_size_not_divisible(
        batch_size=config.train_batch_size_per_learner,
        num_envs_per_env_runner=config.num_envs_per_env_runner,
    )
    warn_if_minibatch_size_not_divisible(
        minibatch_size=config.minibatch_size,
        num_envs_per_env_runner=config.num_envs_per_env_runner,
    )
    config.validate_train_batch_size_vs_rollout_fragment_length()
    return config, module_spec
